!******************************************************************************
!
! $Id: atlas.F90,v 1.1 2010/07/10 01:42:42 murakami Exp $
!
!******************************************************************************
!==============================================================================
!  From book by William J.Thompsom: 
!  "ATLAS FOR COMPUTING MATHEMATICAL FUNCTIONS"
!  Chap.12, Elliptic integrals and elliptic functions".
!  page 499.
!---
!  Elliptic integrals of first kind F(phi|m) and K(m) by AGM method; 
!  (Note, the argument m is in the traditional mathematical
!  formulae denoted as k^2.)
!==============================================================================
FUNCTION ATLAS_ELLPTC_INTGRL_FK (PHI, M, EPS, KM) RESULT(VAL)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK) VAL
REAL(RK),INTENT(IN):: PHI
REAL(RK),INTENT(IN):: M
REAL(RK),INTENT(IN):: EPS
REAL(RK),INTENT(OUT):: KM
!--------------------------
REAL(RK),PARAMETER:: PID2 = 2*ATAN(1.0_RK)
REAL(RK) A, B, PSI, C, EPSD2, TANPSI, D, TEMP
INTEGER N
!-------------------
  A = 1.0_RK
  B = SQRT(1.0_RK-M)
  PSI = PHI
  C = 1.0_RK
  TEMP = 1.0_RK
  EPSD2 = 0.5_RK * EPS
  N = 0
  DO WHILE (ABS(C) > EPSD2 * TEMP)
    TANPSI = TAN(PSI)
    D = ATAN((A-B)*TANPSI / (A + B * TANPSI*TANPSI))
    PSI = 2.0_RK * PSI - D
    TEMP = A
    A = 0.5_RK * (A + B)
    C = 0.5_RK * (TEMP - B)
    B = SQRT(TEMP * B)
    N = N + 1
  ENDDO
  KM = PID2 / A
  VAL = PSI / ((2.0_RK**N) * A)
END FUNCTION ATLAS_ELLPTC_INTGRL_FK
!==============================================================================
!  From book by William J.Thompsom:
!  "ATLAS FOR COMPUTING MATHEMATICAL FUNCTIONS"
!  Chap.12, Elliptic integrals and elliptic functions", page 521.
!  (Note, the argument m is in the traditional mathematical
!  formulae denoted as "k^2".)
!  JACOBI ELLIPTIC FUNCTIONS BY AGM METHOD; "SN', "CN" AND "DN".
!==============================================================================
SUBROUTINE ATLAS_JACOBI_SNCNDN (U, M, EPS, SN, CN, DN)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: U
REAL(RK),INTENT(IN):: M
REAL(RK),INTENT(IN):: EPS
REAL(RK),INTENT(OUT):: SN, CN, DN
!----------------------------------------
! INTRINSIC ASIN ! INVERSE SINE FUNCTION.
REAL(RK),ALLOCATABLE:: A(:), B(:), C(:), PHI(:)
INTEGER IFLAG ! STATUS FLAG FOR ALLOCATION.
INTEGER N
INTEGER MAX
!----------------------------------------
  SELECT CASE(RK)
    CASE(KIND(1.E0)); MAX=10 ! SINGLE PRECISION.
    CASE(KIND(1.D0)); MAX=10 ! DOUBLE PRECISION.
#ifndef NO_QUADRUPLE_PRECISION
    CASE(KIND(1.Q0)); MAX=20 ! QUADRUPLE PRECISION.
#endif
    CASE DEFAULT; STOP 'ATLAS_JACOBI_SNCNDN:ERROR UNKNOWN KIND.'
  END SELECT
  !--
  ALLOCATE (A(0:MAX-1),B(0:MAX-1),C(0:MAX-1), PHI(0:MAX-1), STAT=IFLAG)
  IF (IFLAG/=0) STOP 'ATLAS_JACOBI_SNCNDN: FAILED TO ALLOCATE ARRAYS.'
  !--
  IF(M == 0.0_RK) THEN
    SN = SIN(U)
    CN = COS(U)
    DN = 1.0_RK
    RETURN
  ENDIF
  IF(M == 1.0_RK) THEN
    SN = TANH(U)
    CN = 1.0_RK / COSH(U)
    DN = CN
    RETURN
  END IF
  A(0) = 1.0_RK
  B(0) = SQRT(1.0_RK - M)
  C(0) = SQRT(M)
  N = 0
  DO WHILE (ABS(C(N)) > EPS)
    A(N+1) = 0.5_RK * (A(N) + B(N))
    B(N+1) = SQRT(A(N) * B(N))
    C(N+1) = 0.5_RK * (A(N) - B(N))
    IF (N  >= MAX-2) THEN
      WRITE(*,*) '!!IN JACOBI_SNCNDN N=", N, " TOO LARGE'
      WRITE(*,*) 'INCREASE "MAX" OR DECREASE EPS.'
#if 1
      STOP 'ATLAS_JACOBI_SNCNDN: STOPPED.'
#endif
      C(MAX-1) = 0.0_RK
    ENDIF
    N = N + 1
  ENDDO
  N = N - 1
  PHI(N) = (2.0_RK**N) * A(N) * U
  DO WHILE (N > 0)
    PHI(N-1) = 0.5_RK * (PHI(N) + ASIN(C(N) * SIN(PHI(N)) / A(N)))
    N = N - 1
  ENDDO
  SN = SIN(PHI(0))
  CN = COS(PHI(0))
  DN = CN / COS(PHI(1) - PHI(0))
  !--
  DEALLOCATE (A, B, C, PHI)
END SUBROUTINE ATLAS_JACOBI_SNCNDN
!============================================================================
!  Calculate the elliptic integrals of the first kind
!  "F(phi|1-m)" and "K(1-m)" for the small "m" in magnitude.
!  (Note, the argument "m" is in the traditional mathematical formulae 
!  denoted as "k^2".)
!
!  This is a MODIFIED routine from the elliptic integrals of first kind
!  "F(phi|m)" and "K(m)" by AGM method. From book by William J.Thompsom: 
!  "ATLAS FOR COMPUTING MATHEMATICAL FUNCTIONS", 
!  Chap.12, Elliptic integrals and elliptic functions", page 499.
!============================================================================
FUNCTION ATLAS_ELLPTC_INTGRL_FK1 (PHI, M, EPS, KM) RESULT(VAL)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK) VAL
REAL(RK),INTENT(IN):: PHI
REAL(RK),INTENT(IN):: M
REAL(RK),INTENT(IN):: EPS
REAL(RK),INTENT(OUT):: KM
!------------------------------------------------
REAL(RK),PARAMETER:: PID2 = 2*ATAN(1.0_RK) ! PI/2
REAL(RK) A, B, PSI, C, EPSD2, TANPSI, D, TEMP
INTEGER N
!-------------------------------------------------
  A = 1.0_RK
  B = SQRT(M) ! THIS LINE IS MODIFIED FROM THE ORIGINAL IN THE BOOK.
  PSI = PHI
  C = 1.0_RK
  TEMP = 1.0_RK
  EPSD2 = 0.5_RK * EPS
  N = 0
  DO WHILE (ABS(C) > EPSD2 * TEMP)
    TANPSI = TAN(PSI)
    D = ATAN((A-B) * TANPSI / (A + B * TANPSI * TANPSI))
    PSI = 2.0_RK * PSI - D
    TEMP = A
    A = 0.5_RK * (A + B)
    C = 0.5_RK * (TEMP - B)
    B = SQRT(TEMP * B)
    N = N + 1
  ENDDO
  KM = PID2 / A
  VAL = PSI / ((2.0_RK**N) * A)
END FUNCTION ATLAS_ELLPTC_INTGRL_FK1
