!******************************************************************************
!
! $Id: filter_ellptc.F90,v 1.1 2010/07/10 01:42:42 murakami Exp $
!
!******************************************************************************
!==============================================================================
! CALCULATES THE VALUE OF "A_STOP", THE TRUE MINIMUM ATTENUATION IN THE STOP 
! BAND OF THE ELLIPTIC FILTER.
!==============================================================================
SUBROUTINE FILTER_ASTOP_ELLPTC (EPS, L, A_STOP)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN) :: EPS, L ! "EPS" AND "L" IN THE FORMULA OF ATTENUATION.
REAL(RK),INTENT(OUT):: A_STOP ! THE VALUE OF "A_STOP".
!----------------------------
  A_STOP = 1 + (EPS*L)**2
END SUBROUTINE FILTER_ASTOP_ELLPTC
!===============================================================================
!
!
!===============================================================================
MODULE MOD_ELLPTC
IMPLICIT NONE
INTEGER,PRIVATE,PARAMETER:: RK = MY_RKIND
REAL(RK),PARAMETER:: TOL = EPSILON(1.0_RK)
REAL(RK),PRIVATE:: K   ! KEEPS THE VALUE OF "1/MU" IN THIS VARIABLE.
REAL(RK),PRIVATE:: KK   ! KEEPS THE VALUE OF ELLIPTIC COMPLETE INTEGRAL OF "K".
REAL(RK),PRIVATE:: KK_1 ! KEEPS THE VALUE OF ELLIPTIC COMPLETE INTEGRAL OF "K'",
                       ! WHERE "K'=SQRT(1-K^2)".
REAL(RK),PRIVATE:: C   ! COEFFICIENT OF "R_N"
REAL(RK),PRIVATE,ALLOCATABLE:: TP(:), TZ(:) ! POLES AND ZEROS OF "R_N(T)".
PRIVATE:: ELLPTC_CMPLT_INTGRL, ELLPTC_SN
REAL(RK),EXTERNAL:: ATLAS_ELLPTC_INTGRL_FK1 ! ATLAS LIBRARY FUNCTION.
CONTAINS
!===============================================================================
!   CALCULATES THE VALUES OF "EPS" AND "N_MIN" OF THE ELLIPTIC FILTER
!   FROM THE VALUES OF "A_MAX", "A_MIN" AND "MU".
!===============================================================================
SUBROUTINE FILTER_SETUP_ELLPTC (MU, A_MAX, A_MIN, EPS, N_MIN)
REAL(RK),INTENT(IN):: MU, A_MAX, A_MIN ! THREE DESIGN PARAMETERS OF ATTENUATION.
REAL(RK),INTENT(OUT):: EPS  ! VALUE OF "EPS" IN THE FORMULA OF ATTENUATION.
INTEGER,INTENT(OUT):: N_MIN ! MINIMUM DEGREE TO SATISFY THE DESIGN.
!-------------------------------------------------------------
REAL(RK) L_MIN
REAL(RK) KK_L, KK_L1 ! VARIABLES KK_L AND KK_L1 ARE LOCAL TO THIS ROUTINE.
!-----------------------------------
  EPS = SQRT(A_MAX-1)
  L_MIN = SQRT((A_MIN-1)/(A_MAX-1))
  !--
  K = 1 / MU
  KK   = ELLPTC_CMPLT_INTGRL(1/MU)
  KK_1 = ELLPTC_CMPLT_INTGRL(SQRT((1-1/MU)*(1+1/MU)))
  !--
  KK_L  = ELLPTC_CMPLT_INTGRL(1/L_MIN)
  KK_L1 = ELLPTC_CMPLT_INTGRL(SQRT((1-1/L_MIN)*(1+1/L_MIN)))
  ! CALCULATION OF THE VALUE OF N_MIN, THE LOWER BOUND FOR N.
  N_MIN = CEILING((KK_L1/KK_L) / (KK_1/KK))
END SUBROUTINE FILTER_SETUP_ELLPTC
!===============================================================================
! CALCULATES THE VALUE OF "L" FROM THE VALUES OF "MU" AND "N".
!===============================================================================
SUBROUTINE FILTER_SETUP_ELLPTC2 (MU, N, L)
REAL(RK),INTENT(IN):: MU  ! PARAMETER "MU" OF THE SHAPE OF ATTENUATION.
INTEGER, INTENT(IN):: N   ! DEGREE OF THE FILTER.
REAL(RK),INTENT(OUT):: L  ! VALUE OF "L" IN THE FORMULA OF ATTENUATION.
!-----------------------------------
REAL(RK) VAL
INTEGER J
!-----------------------------------
! CALCULATION OF THE VALUE OF "L" BY THE STANDARD FORMULA.
  VAL = 1.0_RK
  DO  J = 1, N/2
    VAL = VAL * ELLPTC_SN((2*J-1)*KK/N)
  ENDDO
  L = MU**N / VAL**4
END SUBROUTINE FILTER_SETUP_ELLPTC2
!===============================================================================
! CALCULATES THE POLES AND THEIR COEFFICIENTS OF THE ELLIPTIC FILTER.
!===============================================================================
SUBROUTINE FILTER_POLES_COEFS_ELLPTC (N, MU, EPS, L, ZPOLE, ZCOEF, CINF)
INTEGER,    INTENT(IN):: N   ! THE DEGREE OF THE FILTER.
REAL(RK),INTENT(IN):: MU, EPS, L ! "MU","EPS","L" IN THE FORMULA OF ATTENUATION.
COMPLEX(RK),INTENT(OUT):: ZPOLE(N) ! POLES WITH POSITIVE IMAGINARY PARTS,
COMPLEX(RK),INTENT(OUT):: ZCOEF(N) ! AND THEIR COEFFICIENTS.
REAL(RK),   INTENT(OUT):: CINF     ! COEFFICIENT OF THE CONSTANT TERM.
!------------------------------------------
REAL(RK) KK_L, KK_MU
REAL(RK) B, DUMMY, TAU, THETA, RES, MAXRES
REAL(RK) SNX, CNX, DNX, SNY1, CNY1, DNY1, DEN, VALX, VALY
REAL(RK) TMP
COMPLEX(RK) Z1, Z2
INTEGER I, J
!--------------------------
! CALCULATES THE POLES OF THE TRANSFER FUNCTION (ZEROS OF ATTENUATION FUNCTION).
  ! "$b := F(\atan{1/\eps},k')$";
  B = ATLAS_ELLPTC_INTGRL_FK1(ATAN2(1.0_RK,EPS), (1/L)**2, TOL, DUMMY)
  ! CALCULATES THE VALUE OF "TAU" WHICH IS COMMON TO ALL POLES.
  KK_L  = ELLPTC_CMPLT_INTGRL(1/L)
  KK_MU = ELLPTC_CMPLT_INTGRL(1/MU)
  TAU = (KK_MU/KK_L) * (B/N)
  ! CALCULATES THE VALUES OF ELLIPTIC FUNCTIONS "SN(Y), CN(Y), DN(Y)" 
  ! FOR A REAL ARGUMENT "Y=TAU".
  CALL ATLAS_JACOBI_SNCNDN (TAU, (1-1/MU)*(1+1/MU), TOL, SNY1, CNY1, DNY1) 
  DO  J = 0, 2*N-1
    I = 1+MOD(J+N/2,2*N)
    IF (I > N) CYCLE ! THOSE POLES WITH NEGATIVE IMAGINARY PARTS ARE AVOIDED.
    THETA = (2*J + (1-MOD(N,2))) * (KK_MU/N) 
    ! CALCULATES THE VALUES OF ELLIPTIC FUNCTIONS "SN(X), CN(X), DN(X)" 
    ! FOR A REAL ARGUMENT "X=THETA".
    CALL ATLAS_JACOBI_SNCNDN (THETA, (1/MU)**2, TOL, SNX,  CNX,  DNX) 
    ! "(VALX,VALY) := COMPLEX_SN((X,Y),1/MU)";
    ! THE REAL AND IMAGINARY PARTS OF ELLIPTIC FUNCTION "SN" WITH COMPLEX
    ! ARGUMENT "Z=(X,Y)" CAN BE REPRESENTED BY A RATIONAL EXPRESSION OF 
    ! ELLIPTIC FUNCTIONS "SN", "CN", "DN" FOR REAL ARGUMENTS "X" AND "Y".
    DEN = CNY1**2 + (K * SNX * SNY1)**2
    VALX = (SNX * DNY1) / DEN
    VALY = (CNX * DNX * SNY1 * CNY1) / DEN
    ZPOLE(I) = CMPLX(VALX, VALY, KIND=RK)
  ENDDO
#define ASSERT
#ifdef ASSERT
  ! CHECK WHETHER ALL CALCULATED POLES HAVE POSITIVE IMAGINARY PARTS.
  DO  I = 1, N
    IF (AIMAG(ZPOLE(I)) <= 0.0_RK) THEN
      PRINT *, 'MAKE_POLES_OF_AINV: DETECTED NON-POSITIVE IMAGINARY PART.'
      STOP
    ENDIF
  ENDDO
#endif
!------------------------------------------
! MAKE THE COEFFICIENTS "ZCOEF" OF THE POLES.
! (NOTE: THIS CODE IS NOT FULLY OPTIMIZED. SINCE TAU IS INVARIANT 
! INSIDE THE DO LOOP, CALCULATION CAN BE ECHONOMIZED BY USING 
! ELLIPTIC FUNCTIONS OF REAL ARGUMENTS ONLY.)
  TMP = (KK_MU/KK_L) / (2*N) * EPS/SQRT((1+EPS**2)*(EPS**2+(1/L)**2))
  DO  J = 0, 2*N-1
    I = 1+MOD(J+N/2,2*N)
    IF (I > N) CYCLE ! SKIP CALCULATION FOR POLES WITH NEGATIVE IMAGINARY PARTS.
    THETA = (2*J + (1-MOD(N,2))) * (KK_MU/N) 
    CALL CMPLX_ELLPTC_CN (THETA, TAU, 1/MU, VALX, VALY)
    Z1 = CMPLX(VALX, VALY, KIND=RK)
    CALL CMPLX_ELLPTC_DN (THETA, TAU, 1/MU, VALX, VALY)
    Z2 = CMPLX(VALX, VALY, KIND=RK)
    ZCOEF(I) = CMPLX(0,-1,KIND=RK) * TMP * Z1 * Z2
  ENDDO
!-----------------------------------------------
! SET "CINF" THE VALUE OF COEFFICIENT OF THE CONSTANT TERM.
  IF (MOD(N,2) == 0) THEN
    CINF = 1 / (1 + (EPS*L)**2)
  ELSE
    CINF = 0.0_RK
  ENDIF
!------
END SUBROUTINE FILTER_POLES_COEFS_ELLPTC
!===============================================================================
! CALCULATION OF THE COMPLETE ELLIPTIC INTEGRAL OF THE FIRST KIND.
!===============================================================================
FUNCTION ELLPTC_CMPLT_INTGRL(K) RESULT(VAL)
REAL(RK) VAL
REAL(RK),INTENT(IN):: K ! MODULUS.
!----------------------------------------------
REAL(RK),PARAMETER:: PI = 4*ATAN(1.0_RK)
REAL(RK) A, B, P, X, Y, E
INTEGER I, J
!--------------------
  A = ATAN2(K, SQRT((1-K)*(1+K)))
  B = PI / 2
  I = 0
  P = 1.0_RK
  DO 
    X = 2 / (1 + SIN(A)) - 1
    Y = SIN(A) * SIN(B)
    A = ATAN2(SQRT((1-X)*(1+X)), X)
    B = (B + ATAN2(Y, SQRT((1-Y)*(1+Y)))) / 2
    P = P * (1 + COS(A))
    E = 1 - A * (2 / PI)
    I = I + 1
    IF (E < TOL) EXIT
  ENDDO 
  X = (PI / 4) + B / 2
  VAL = LOG(TAN(X)) * P
END FUNCTION ELLPTC_CMPLT_INTGRL
!===============================================================================
!   CALCULATION OF "SN(U,K)", THE JACOBI ELLIPTIC FUNCTION "SN".
!   (THE VALUE "K=1/MU" IS HOLD IN THE VARIABLE "K" IN THE MODULE.)
!===============================================================================
FUNCTION ELLPTC_SN (U) RESULT(VAL)
REAL(RK) VAL
REAL(RK),INTENT(IN):: U  ! REAL ARGUMENT.
! REAL VARIABLES k=1/MU, KK=K(1/MU), KK_1=K'(1/MU)=K(k') ARE IN THE MODULE.
!----------------------------------------------------------------------------
REAL(RK),PARAMETER:: PI = 4*ATAN(1.0_RK)
REAL(RK) Q, V, W
INTEGER J
!-----------------
  Q = EXP(-PI*KK_1/KK)
  V = (PI/2)*U/KK
  J = 0
  W = SQRT(Q)
  VAL = 0.0_RK
  DO
    VAL = VAL + W * SIN((2*J+1)*V) / ((1-W)*(1+W))
    J = J + 1
    IF (W < TOL) EXIT
    W = W * Q
  ENDDO
  VAL = VAL * (2*PI) / (K * KK)
END FUNCTION ELLPTC_SN
!===============================================================================
!  CALCULATES THE POLES TP(1:N/2), THE ZEROS TZ(1:N/2), AND THE CONSTANT C.
!===============================================================================
SUBROUTINE SET_RN_POLES_ZEROS (N, MU)
INTEGER, INTENT(IN):: N  ! DEGREE OF RATIONAL FUNCTION.
REAL(RK),INTENT(IN):: MU ! SHAPE PARAMETER "MU".
!--------------------------------------
INTEGER NI, J
REAL(RK) VAL
!----------------------------------
  IF (ALLOCATED(TZ)) DEALLOCATE(TZ)
  IF (ALLOCATED(TP)) DEALLOCATE(TP)
  ALLOCATE (TZ(N/2), TP(N/2))
  NI = MOD(N+1,2)
  VAL = 1.0_RK
  DO  J = 1, N/2
    TZ(J) = ELLPTC_SN((2*J-NI)*KK/N) ! ZEROS OF RN(X) IN ASCENDING ORDER.
    TP(J) = MU / TZ(J)               ! POLES OF RN(X) IN DECENDING ORDER.
    VAL = VAL * ((1-TP(J))*(1+TP(J))) / ((1-TZ(J))*(1+TZ(J)))
  ENDDO
  C = VAL
END SUBROUTINE SET_RN_POLES_ZEROS
!==============================================================================
!  THE ELLIPTIC RATIONAL FUNCTION OF DEGREE "N".
!==============================================================================
FUNCTION ELLPTC_RN (N, T) RESULT(VAL)
REAL(RK) VAL ! THE FUNCTION RETURN VALUE.
INTEGER, INTENT(IN):: N ! DEGREE
REAL(RK),INTENT(IN):: T ! THE NORMAL COORDINATE.
!-------------------------------------------------
INTEGER J
!----------------------------
  IF (MOD(N,2) == 1) THEN
    VAL = C * T ! FOR ODD "N".
  ELSE
    VAL = C     ! FOR EVEN "N".
  ENDIF
  DO  J = 1, N/2
    VAL = VAL * (((T-TZ(J))*(T+TZ(J))) / ((T-TP(J))*(T+TP(J))))
  ENDDO
END FUNCTION ELLPTC_RN
!===============================================================================
!  ELLIPTIC RATIONAL FUNCTION OF DEGREE "N" OF A COMPLEX ARGUMENT.
!===============================================================================
FUNCTION CMPLX_ELLPTC_RN (N, T) RESULT(VAL)
COMPLEX(RK) VAL  ! THE FUNCTION RETURN VALUE.
INTEGER,    INTENT(IN):: N ! THE DEGREE OF RATIONAL FUNCTION.
COMPLEX(RK),INTENT(IN):: T ! THE RELATIVE COORDINATE.
!----------------------------------------------------
INTEGER J
!---------------------------
  IF (MOD(N,2) == 1) THEN
    VAL = C * T ! FOR ODD N.
  ELSE
    VAL = C     ! FOR EVEN N.
  ENDIF
  DO  J = 1, N/2
    VAL = VAL * (((T-TZ(J))*(T+TZ(J))) / ((T-TP(J))*(T+TP(J))))
  ENDDO
END FUNCTION CMPLX_ELLPTC_RN
END MODULE MOD_ELLPTC
!==============================================================================
!  THE ELLIPTIC FUNCTION "SN" WITH COMPLEX ARGUMENT.
!==============================================================================
SUBROUTINE CMPLX_ELLPTC_SN (X, Y, K, VALX, VALY)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: X, Y ! REAL AND IMAGINARY PARTS OF THE ARGUMENT.
REAL(RK),INTENT(IN):: K    ! MODULUS (THE CLASSICAL DEFINITION).
REAL(RK),INTENT(OUT):: VALX, VALY ! REAL AND IMAGINARY PARTS OF THE RESULT.
!--------------------------------------
REAL(RK),PARAMETER:: TOL = EPSILON(1.0_RK)
REAL(RK) K2, DEN
REAL(RK) SNX, CNX, DNX, SNY1, CNY1, DNY1
!--------------------------------------
  K2 = (1-K)*(1+K) ! K2 IS K1**2, WHERE K1 = SQRT((1-K)*(1+K)).
  CALL ATLAS_JACOBI_SNCNDN (X, K**2, TOL, SNX,  CNX,  DNX) 
  CALL ATLAS_JACOBI_SNCNDN (Y, K2, TOL, SNY1, CNY1, DNY1) 
  DEN = CNY1**2 + (K * SNX * SNY1)**2
  VALX = (SNX * DNY1) / DEN
  VALY = (CNX * DNX * SNY1* CNY1) / DEN
END SUBROUTINE CMPLX_ELLPTC_SN
!==============================================================================
! THE ELLIPTIC FUNCTION "CN" WITH COMPLEX ARGUMENT.
!==============================================================================
SUBROUTINE CMPLX_ELLPTC_CN (X, Y, K, VALX, VALY)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: X, Y ! REAL AND IMAGINARY PARTS OF THE ARGUMENT.
REAL(RK),INTENT(IN):: K    ! MODULUS (THE CLASSICAL DEFINITION).
REAL(RK),INTENT(OUT):: VALX, VALY ! REAL AND IMAGINARY PARTS OF THE RESULT.
!--------------------------------------
REAL(RK),PARAMETER:: TOL = EPSILON(1.0_RK)
REAL(RK) K2, DEN
REAL(RK) SNX, CNX, DNX, SNY1, CNY1, DNY1
!--------------------------------------
  K2 = (1-K)*(1+K) ! K2 IS K1**2, WHERE K1 = SQRT((1-K)*(1+K)).
  CALL ATLAS_JACOBI_SNCNDN (X, K**2, TOL, SNX,  CNX,  DNX) 
  CALL ATLAS_JACOBI_SNCNDN (Y, K2, TOL, SNY1, CNY1, DNY1) 
  DEN = CNY1**2 + (K * SNX * SNY1)**2
  VALX = (CNX * CNY1) / DEN
  VALY = -(SNX * DNX * SNY1 * DNY1) / DEN
END SUBROUTINE CMPLX_ELLPTC_CN
!==============================================================================
!  THE ELLIPTIC FUNCTION "DN" WITH COMPLEX ARGUMENT.
!==============================================================================
SUBROUTINE CMPLX_ELLPTC_DN (X, Y, K, VALX, VALY)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: X, Y ! REAL AND IMAGINARY PARTS OF THE ARGUMENT.
REAL(RK),INTENT(IN):: K    ! MODULUS (THE CLASSICAL DEFINITION).
REAL(RK),INTENT(OUT):: VALX, VALY ! REAL AND IMAGINARY PARTS OF THE RESULT.
!--------------------------------------------------------------------------
REAL(RK),PARAMETER:: TOL = EPSILON(1.0_RK)
REAL(RK) K2, DEN
REAL(RK) SNX, CNX, DNX, SNY1, CNY1, DNY1
!--------------------------------------
  K2 = (1-K)*(1+K) ! K2 IS K1**2, WHERE K1 = SQRT((1-K)*(1+K)).
  CALL ATLAS_JACOBI_SNCNDN (X, K**2, TOL, SNX,  CNX,  DNX) 
  CALL ATLAS_JACOBI_SNCNDN (Y, K2, TOL, SNY1, CNY1, DNY1) 
  DEN = CNY1**2 + (K * SNX * SNY1)**2
  VALX = (DNX * CNY1 * DNY1) / DEN
  VALY = (-K**2 * SNX * CNX * SNY1) / DEN
END SUBROUTINE CMPLX_ELLPTC_DN
