!******************************************************************************
!
! $Id: filter_invcheb.F90,v 1.1 2010/07/10 01:42:42 murakami Exp $
!
!******************************************************************************
!==============================================================================
! CALCULATES THE VALUE OF "A_STOP", THE TRUE MINIMUM ATTENUATION IN THE STOP 
! BAND OF THE INVERSE-CHEBYSHEV FILTER.
!==============================================================================
SUBROUTINE FILTER_ASTOP_INVCHEB (EPS, MU, N, A_STOP)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: EPS, MU ! "EPS" AND "MU" FOR THE FORMULA OF ATTENUATION.
INTEGER,INTENT(IN):: N        ! DEGREE OF THE FILTER.
REAL(RK),INTENT(OUT):: A_STOP ! CALCULATED VALUE OF "A_STOP".
!----------------------------
REAL(RK),EXTERNAL:: CHEB
!------------------------
  A_STOP = 1 + EPS**2 * CHEB(N, MU)**2
END SUBROUTINE FILTER_ASTOP_INVCHEB
!===============================================================================
! CALCULATES THE VALUES OF "EPS" AND "N_MIN" OF THE INVERSE-CHEBYSHEV FILTER
! FROM THE VALUES OF "A_MAX", "A_MIN" AND "MU".
!===============================================================================
SUBROUTINE FILTER_SETUP_INVCHEB (MU, A_MAX, A_MIN, EPS, N_MIN)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: MU, A_MAX, A_MIN ! INPUT SHAPE PARAMETERS OF ATTENUATION.
REAL(RK),INTENT(OUT):: EPS   ! PARAMETER "EPS" IN THE FORMULA OF ATTENUATION.
INTEGER, INTENT(OUT):: N_MIN ! THE LEAST VALUE OF N REQUIRED.
!--------------------------------------------------------------
! INTRINSIC:: ACOSH ! INVERSE HYPERBOLIC-COSINE (FORTRAN2008).
REAL(RK) L_MIN
!------------------------------------
  EPS = SQRT(A_MAX-1)
  L_MIN = SQRT((A_MIN-1)/(A_MAX-1))
  N_MIN = CEILING(ACOSH(L_MIN)/ACOSH(MU))
END SUBROUTINE FILTER_SETUP_INVCHEB
!==============================================================================
!  CALCULATES THE POLES AND THEIR COEFFICIENTS OF THE INVERSE-CHEBYSHEV FILTER.
!  THE POLES ARE IN THE STANDARD COORDINATE.
!  NOTE, ONLY THOSE POLES WHOSE IMAGINARY PARTS ARE POSITIVES AND THEIR
!  COEFFICIENTS ARE CALCULATED.
!==============================================================================
SUBROUTINE FILTER_POLES_COEFS_INVCHEB (N, MU, EPS, ZPOLE, ZCOEF, CINF)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN):: N   ! THE DEGREE OF THE FILTER.
REAL(RK),INTENT(IN):: MU  ! THE "MU" IN THE FORMULA OF ATTENUATION.
REAL(RK),INTENT(IN):: EPS ! THE "EPS" IN THE FORMULA OF ATTENUATION.
COMPLEX(RK),INTENT(OUT):: ZPOLE(N) ! POLES WITH POSITIVE IMAGINARY PARTS, 
COMPLEX(RK),INTENT(OUT):: ZCOEF(N) ! AND THEIR COEFFICIENTS, CORRESPONDINGLY.
REAL(RK),   INTENT(OUT):: CINF     ! THE COEFFICIENT FOR THE CONSTANT TERM.
!----------------------------------------------------------------------------
! INTRINSIC ASINH ! INVERSE HYPERBOLIC-SINE; (FORTRAN2008).
REAL(RK),PARAMETER:: PI = 4*ATAN(1.0_RK)
COMPLEX(RK),EXTERNAL:: ZCHEB, ZCHEB2ND
REAL(RK),EXTERNAL:: CHEB
REAL(RK) TAU, THETA, L
COMPLEX(RK) X
INTEGER J
!-------------------------
  ! CALCULATES THOSE POLES WHOSE IMAGINARY PARTS ARE POSITIVES,
  ! AND THEIR COEFFICIENTS.
  L = CHEB(N,MU)
  TAU = ASINH(EPS*L) / N
  DO  J = 1, N
    THETA = (2*J-1)*PI/(2*N)
    X = CMPLX(COSH(TAU)*COS(THETA),-SINH(TAU)*SIN(THETA),KIND=RK)
    ZPOLE(J) = MU/X
    ZCOEF(J) = -MU/(2*N*X**2)*ZCHEB(N,X)/ZCHEB2ND(N-1,X)
  ENDDO
  !-----------------------
  ! CALCULATE "CINF" THE COEFFICIENT OF THE CONSTANT TERM.
  IF (MOD(N,2) == 0) THEN
    CINF = 1/(1+(EPS*L)**2)
  ELSE
    CINF = 0.0_RK
  ENDIF
  !------------
END SUBROUTINE FILTER_POLES_COEFS_INVCHEB
