!******************************************************************************
!
! $Id: b_orth.F90,v 1.4 2010/07/12 12:36:15 murakami Exp $
!
!******************************************************************************
!==============================================================================
!
!   "B"-ORTHONORMALIZATION OF "NV" VECTORS "X" BY THE MODIFIED GRAM-SCHMIDT
!   METHOD. (NOTE: THIS CODE IS SLOW WHEN "NV" IS LARGE, SINCE THE MATRIX "B"
!   IS SCANNED "NV" TIMES).
!
!==============================================================================
SUBROUTINE B_ORTH_MGS (BAND_B, N, NLD, NB, NV, X)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_B(NB,N) ! STORE OF BANDED SYMMETRIC MATRIX "B".
INTEGER, INTENT(IN):: N            ! SIZE OF MATRIX "B" AND VECTOR "X".
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "B".
INTEGER, INTENT(IN):: NB           ! LEADING DIMENSION OF "BAND_B".
INTEGER, INTENT(IN):: NV         ! NUMBER OF VECTORS IN "X".
REAL(RK),INTENT(INOUT):: X(N,NV) ! SET OF VECTORS "X".
!------------------------------------------------------------------------------
REAL(RK),ALLOCATABLE:: Y(:)
INTEGER IFLAG
REAL(RK) R_LL, R_LJ
INTEGER L, J
!---------------------------------
  ALLOCATE (Y(N), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. B_ORTH_MGS: FAILED TO ALLOCATE Y.'
   !---
  DO  L = 1, NV
    CALL BAND_MULT (BAND_B, N, NLD, SIZE(BAND_B,1), X(:N,L), Y(:N))
    R_LL = SQRT(DOT_PRODUCT(X(:N,L), Y(:N)))
    X(:N,L) = X(:N,L) / R_LL
    Y(:N) = Y(:N) / R_LL
    DO  J = L+1, NV
      R_LJ = DOT_PRODUCT(Y(:N), X(:N,J))
      X(:N,J) = X(:N,J) - R_LJ * X(:N,L)
    ENDDO
  ENDDO
  DEALLOCATE (Y)
END SUBROUTINE B_ORTH_MGS
!==============================================================================
!
!    "B"-ORTHONORMALIZATION BY THE CLASSICAL "B"-SVD METHOD.
!
!==============================================================================
SUBROUTINE B_ORTH_SVD (BAND_B, N, NLD, NB, NV, Y, REL_THRES, SIGMA, RANK)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_B(NB,N) ! STORE OF BANDED SYMMETRIC MATRIX "B".
INTEGER, INTENT(IN):: N            ! SIZE OF MATRIX "B" AND VECTORS.
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "B"
INTEGER, INTENT(IN):: NB           ! LEADING DIMENSION OF "BAND_B".
INTEGER, INTENT(IN):: NV         ! NUMBER OF VECTORS IN "Y".
REAL(RK),INTENT(INOUT):: Y(N,NV) ! SET OF VECTORS "Y" TO ORTHONORMALIZE.
REAL(RK),INTENT(IN):: REL_THRES  ! RELATIVE THRESHOLD OF SINGULAR VALUE.
REAL(RK),INTENT(OUT):: SIGMA(NV) ! SINGULAR VALUES.
INTEGER, INTENT(OUT):: RANK      ! EFFECTIVE RANK DETERMINED BY SVD.
!------------------------------------------------------------------------------
INTEGER,PARAMETER:: MX_LP = 2 ! TWICE FOR ITERATION SEEMS ENOUGH.
LOGICAL,PARAMETER:: SHOW_SVD_SIGMA = .TRUE.
INTEGER,PARAMETER:: FILE_66 = 66
INTEGER,PARAMETER:: FILE_88 = 88 ! DATA FILE OF SINGULAR VALUES FOR GNUPLOT.
CHARACTER(*),PARAMETER:: FNAME_88 = 'sigma.data'
REAL(RK),ALLOCATABLE:: S(:,:), U(:,:), W(:,:)
INTEGER IFLAG
INTEGER LP, L, NROT
REAL(RK) ERR
!---------------------------------
  ALLOCATE (S(NV,NV), U(NV,NV), W(NV,NV), STAT=IFLAG)
  IF (IFLAG/=0) STOP 'B_ORTH_SVD: FAILED TO ALLOCATE S, U, W.'
  !--
  LOOP: DO  LP = 1, MX_LP
    !=============================
    ! MAKE THE PROJECTED SMALL MATRIX : "S := Y^{T} * B * Y".
    CALL MAKE_PROJECTED_MAT (BAND_B, N, NLD, SIZE(BAND_B,1), &
                         NV, Y, SIZE(S,1), S)
    !=============================
    ! EIGENVALUE DECOMPOSITION OF "S" TO MAKE SQUARE OF SINGULAR VALUES "SIGMA"
    ! AND THE SET OF SINGULAR VECTORS "U".
    CALL RSYM_JACOBI (NV, S, SIZE(S,1), SIGMA, U, SIZE(U,1), NROT)
#if 1
    PRINT 100, 'B_ORTH_SVD LP=', LP, ', RSYM_JACOBI NROT=', NROT
    100 FORMAT (A, I1, A, I10)
#endif
    !=============================
    ! SINGULAR PAIRS ARE SORTED IN THEIR SINGULAR VALUES IN DESCENDING ORDER.
    CALL EIG_SORT (NV, SIGMA, U, SIZE(U,1), W, SIZE(W,1), 'DESCENDING')
    !=============================
    ! THE SET OF VECTORS "Y" ARE ROTATED BY "U" TO MAKE AN ORTHOGONAL BASIS SET.
#if 0
    Y(:N,:NV) = MATMUL(Y(:N,:NV), U(:NV,:NV))
#else
    CALL RIGHT_TRFM (N, NV, Y, SIZE(Y,1), U, SIZE(U,1))
#endif
    !=============================
  ENDDO LOOP
  !===============================
  ! SINGULAR VALUES ARE CALCULATED FROM THEIR SQUARES.
  DO  L = 1, NV
    IF (SIGMA(L) > 0.0_RK) THEN
      SIGMA(L) = SQRT(SIGMA(L))
    ELSE
      SIGMA(L) = 0.0_RK
    ENDIF
  ENDDO
  !===============================
  ! THE EFFECTIVE RANK IS DETERMINED BY THE RELATIVE THRESHOLD "REL_THRES".
  RANK = 0
  IF (SIGMA(1) > 0.0_RK) THEN
    DO  L = 1, NV
      IF (SIGMA(L) <= SIGMA(1) * REL_THRES) EXIT
      RANK = RANK + 1
    ENDDO
  ENDIF
  !===============================
  ! NORMALIZE SINGULAR VECTORS WHOSE SINGULAR VALUES ARE ABOVE THE THRESHOLD.
  DO  L = 1, RANK
    Y(:N,L) = Y(:N,L) / SIGMA(L)
  ENDDO
  !===============================
  IF (SHOW_SVD_SIGMA) THEN
    ! SHOW THE SINGULAR VALUES AND THE DETERMINED EFFECTIVE RANK.
    WRITE(      *,*) 'SINGULAR VALUES:'
    WRITE(FILE_66,*) 'SINGULAR VALUES:'
    DO  L = 1, NV
      WRITE(      *,200) L, SIGMA(L)
      WRITE(FILE_66,200) L, SIGMA(L)
      200 FORMAT(3X, 'SIGMA(', I4, ') =', ES9.1)
    ENDDO
    WRITE(      *,300) RANK
    WRITE(FILE_66,300) RANK
    300 FORMAT('RANK_SVD=', I4)
  ENDIF
  !===============================
  ! WRITE FILE_88 THE DATA OF SINGULAR VALUES AND THE THRESHOLD FOR GNUPLOT.
  OPEN(FILE_88,FILE=FNAME_88, FORM='FORMATTED')
  DO  L = 1, NV
    WRITE(FILE_88, FMT=310) L, SIGMA(L)
    310 FORMAT(I3, 3X, ES9.1)
  ENDDO
  WRITE(FILE_88,'(//)')
  WRITE(FILE_88, FMT=310)  1, SIGMA(1)*REL_THRES
  WRITE(FILE_88, FMT=310) NV, SIGMA(1)*REL_THRES
  CLOSE(FILE_88)
  !===============================
#define DEBUG
#ifdef DEBUG
  ! CHECK WHETHER THE CALCULATED SINGULAR VECTORS FORM AN ORTHONORMAL SET.
  CALL MAKE_PROJECTED_MAT (BAND_B, N, NLD, SIZE(BAND_B,1), &
                       RANK, Y, SIZE(S,1), S)
  ERR = 0.0_RK
  DO  L = 1, RANK
    IF (L > 1) ERR = MAX(ERR, MAXVAL(ABS(S(1:L-1,L))))
    ERR = MAX(ERR, ABS(S(L,L)-1.0_RK))
  ENDDO
  WRITE(      *,400) ERR 
  WRITE(FILE_66,400) ERR
  400 FORMAT('** ORTHOGONAL ERROR OF SINGULAR VECTORS=', ES8.1)
#endif
  !===============================
  DEALLOCATE (S, U, W)
END SUBROUTINE B_ORTH_SVD
