!******************************************************************************
!
! $Id: band.F90,v 1.1 2010/07/10 01:44:28 murakami Exp $
!
!******************************************************************************
!==============================================================================
!
!   COMPUTES THE MATRIX-VECTOR PRODUCT "Y := A * X", WHERE "A" IS A BANDED 
!   REAL SYMMETRIC MATRIX.
!
!==============================================================================
SUBROUTINE BAND_MULT (BAND_A, N, NLD, NA, X, Y)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A"(NOT CHANGED).
INTEGER, INTENT(IN):: N            ! SIZE OF MATRIX "A".
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "A".
INTEGER, INTENT(IN):: NA           ! LEADING DIMENSION OF "BAND_A".
REAL(RK),INTENT(IN) :: X(N) ! INPUT VECTOR.
REAL(RK),INTENT(OUT):: Y(N) ! OUTPUT VECTOR.
!-----------------------------------------------------------------------
REAL(RK) S
INTEGER I, J, M
#define A(I,J) BAND_A(M-(I)+(J),(I))
!------------------------------------
  M = NLD + 1
  Y(:N) = 0.0_RK
  DO  I = 1, N
    DO  J = MAX(1,I-NLD), I
      S = A(I,J)
      Y(I) = Y(I) + S * X(J)
      IF (I == J) CYCLE !  WHEN DIAGONAL ELEMENT.
      Y(J) = Y(J) + S * X(I)
    ENDDO
  ENDDO
  !----
#undef A
END SUBROUTINE BAND_MULT
!==============================================================================
!
!   COMPUTES THE PRODUCT OF A MATRIX AND A SET OF VECTORS "Y := A * X",
!   WHERE "A" IS A BANDED REAL SYMMETRIC MATRIX, AND BOTH "X" AND "Y" ARE 
!   THE SETS OF "NV" VECTORS.
!
!==============================================================================
SUBROUTINE BAND_MULT_NV (BAND_A, N, NLD, NA, NV, X, Y)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A"(NOT CHANGED).
INTEGER, INTENT(IN):: N            ! SIZE OF MATRIX "A".
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "A".
INTEGER, INTENT(IN):: NA           ! LEADING DIMENSION OF "BAND_A".
INTEGER, INTENT(IN) :: NV      ! NUMBER OF SIMULTANEOUS VECTORS.
REAL(RK),INTENT(IN) :: X(N,NV) ! SET OF INPUT VECTORS.
REAL(RK),INTENT(OUT):: Y(N,NV) ! SET OF OUTPUT VECTORS.
!-----------------------------------------------------------------------
INTEGER I, J, M
REAL(RK) S
#define A(I,J) BAND_A(M-(I)+(J),(I))
!------------------------------------
  M = NLD + 1
  Y(:N,:NV) = 0.0_RK
  DO  I = 1, N
    DO  J = MAX(1, I-NLD), I
      S = A(I,J)
      Y(I,:NV) = Y(I,:NV) + S * X(J,:NV)
      IF (I == J) CYCLE !  WHEN DIAGONAL ELEMENT.
      Y(J,:NV) = Y(J,:NV) + S * X(I,:NV)
    ENDDO
  ENDDO
  !----
#undef A
END SUBROUTINE BAND_MULT_NV
!==============================================================================
!
!   BUILDS THE COMPLEX SHIFTED MATRIX "C := A - MU * B", WHERE BOTH "A" 
!   AND "B" ARE BANDED REAL SYMMETRIC, SHIFT PARAMETER "MU" IS COMPLEX,
!   AND "C" IS COMPLEX SYMMETRIC.
!
!==============================================================================
SUBROUTINE BAND_SHIFT_C (BAND_A, BAND_B, N, NLD, NA, NB, MU, &
                         BAND_C, NC, CNORM)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A"(NOT CHANGED).
REAL(RK),INTENT(IN):: BAND_B(NB,N) ! BANDED SYMMETRIC MATRIX "B"(NOT CHANGED).
INTEGER, INTENT(IN):: N            ! SIZE OF MATRICES "A", "B" AND "C".
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "A", "B" AND "C".
INTEGER, INTENT(IN):: NA           ! LEADING DIMENSION OF "BAND_A".
INTEGER, INTENT(IN):: NB           ! LEADING DIMENSION OF "BAND_B".
COMPLEX(RK),INTENT(IN) :: MU       ! SHIFT PARAMETER IN COMPLEX NUMBER.
COMPLEX(RK),INTENT(OUT):: BAND_C(NC,N) ! SHIFTED BANDED MATRIX "C"("C:=A-MU*B").
INTEGER,    INTENT(IN) :: NC           ! LEADING DIMENSION OF BAND_C.
REAL(RK),   INTENT(OUT):: CNORM        ! NORM OF MATRIX "C".
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: STDERR = 0 ! LOGICAL UNIT NUMBER OF STDERR FOR UNIX.
COMPLEX(RK) W
REAL(RK) S
INTEGER I, J, M
100 FORMAT('SUBR. BAND_SHIFT_C WAS CALLED WITH', 1X, A, I5)
200 FORMAT('SUBR. BAND_SHIFT_C WAS CALLED WITH', 1X, A, I6, A, I6)
#define A(I,J) BAND_A(M-(I)+(J),(I))
#define B(I,J) BAND_B(M-(I)+(J),(I))
#define C(I,J) BAND_C(M-(I)+(J),(I))
!------------------------------------
  M = NLD + 1
  !-- CHECK ARGUMENTS.
  IF (N <= 2) THEN
    WRITE(STDERR, 100) 'N SMALLER THAN 3: N=', N
    STOP 'ERROR: BAND_SHIFT_C.'
  ELSEIF (NA < NLD + 1) THEN
    WRITE(STDERR, 200) 'NA=', NA, ' < NLD+1=', NLD+1
    STOP 'ERROR: BAND_SHIFT_C.'
  ELSEIF (NLD >= N) THEN
    WRITE(STDERR, 200) 'NLD=', NLD, ' >= N=', N
    STOP 'ERROR: BAND_SHIFT_C.'
  ELSEIF (NC <= 3 * NLD) THEN
    WRITE(STDERR, 200) 'NC=', NC, ' <= 3*NLD=', 3*NLD
    STOP 'ERROR: BAND_SHIFT_C.'
  ENDIF
  !---------------------
  CNORM = 0.0_RK
  DO  I = 1, N
    S = 0.0_RK
    DO  J = MAX(1, I-NLD), I
      W = A(I,J) - MU * B(I,J)
      S = S + ABS(W)
      C(I,J) = W
      C(J,I) = W
    ENDDO
    DO  J = I+1, MIN(I+NLD, N)
      S = S + ABS(A(J,I) - MU * B(J,I))
    ENDDO
    DO  J = (I+NLD)+1, MIN(I+2*NLD, N)
      C(I,J) = 0.0_RK
    ENDDO
    CNORM = MAX(CNORM, S)
  ENDDO
  IF (CNORM == 0.0_RK) THEN
    STOP 'SUBR. BAND_SHIFT_C: MATRIX C IS A ZERO MATRIX.'
  ENDIF
  !-----------------------------------------------------
#undef A
#undef B
#undef C
END SUBROUTINE BAND_SHIFT_C
!==============================================================================
!
!   BUILDS THE REAL SHIFTED MATRIX "C := A - MU * B", WHERE BOTH MATRICES 
!   "A" AND "B" ARE BANDED REAL SYMMETRIC, THE SHIFT PARAMETER "MU" IS REAL,
!   AND THE MATRIX "C" IS REAL SYMMETRIC.
!
!==============================================================================
SUBROUTINE BAND_SHIFT (BAND_A, BAND_B, N, NLD, NA, NB, MU, &
                       BAND_C, NC, CNORM)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A"(NOT CHANGED).
REAL(RK),INTENT(IN):: BAND_B(NB,N) ! BANDED SYMMETRIC MATRIX "B"(NOT CHANGED).
INTEGER, INTENT(IN):: N            ! SIZE OF MATRICES "A", "B" AND "C".
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "A", "B" AND "C".
INTEGER, INTENT(IN):: NA           ! LEADING DIMENSION OF "BAND_A".
INTEGER, INTENT(IN):: NB           ! LEADING DIMENSION OF "BAND_B".
REAL(RK),INTENT(IN):: MU           ! SHIFT PARAMETER IN REAL NUMBER.
REAL(RK),INTENT(OUT):: BAND_C(NC,N)! SHIFTED BANDED MATRIX "C" ("C:=A-MU*B").
INTEGER, INTENT(IN):: NC           ! LEADING DIMENSION OF "BAND_C".
REAL(RK),INTENT(OUT):: CNORM       ! NORM OF MATRIX "C".
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: STDERR = 0  ! LOGICAL UNIT NUMBER FOR STDERR FOR UNIX.
REAL(RK) CMX, T, S
INTEGER I, J, K, M
100 FORMAT('SUBR. BAND_SHIFT WAS CALLED WITH', 1X, A, I5)
200 FORMAT('SUBR. BAND_SHIFT WAS CALLED WITH', 1X, A, I6, A, I6)
#define A(I,J) BAND_A(M-(I)+(J),(I))
#define B(I,J) BAND_B(M-(I)+(J),(I))
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-------------------------------------
  ! CHECK ARGUMENTS.
  IF (N <= 2) THEN
    WRITE(STDERR,100) 'N SMALLER THAN 3: N=', N
    STOP 'ERROR: BAND_SHIFT.'
  ELSEIF (NA < NLD + 1) THEN
    WRITE(STDERR,200) 'NA=', NA, ' < NLD+1=', NLD+1
    STOP 'ERROR: BAND_SHIFT.'
  ELSEIF (NLD >= N) THEN
    WRITE(STDERR,200) 'NLD=', NLD, ' >= N=', N
    STOP 'ERROR: BAND_SHIFT.'
  ELSEIF (NC <= 3 * NLD) THEN
    WRITE(STDERR,200) 'NC=', NC, ' <= 3*NLD=', 3*NLD
    STOP 'ERROR: BAND_SHIFT.'
  ENDIF
  !---------------------
  M = NLD + 1
  CNORM = 0.0_RK
  DO  I = 1, N
    S = 0.0_RK
    DO  J = MAX(1,I-NLD), I
      T = A(I,J) - MU * B(I,J)
      C(I,J) = T
      C(J,I) = T
      S = S + ABS(T)
    ENDDO
    DO  J = I+1, MIN(I+NLD,N)
      S = S + ABS(A(J,I) - MU * B(J,I))
    ENDDO
    DO  J = (I+NLD)+1, MIN(I+2*NLD,N)
      C(I,J) = 0.0_RK
    ENDDO
    CNORM = MAX(CNORM,S)
  ENDDO
  IF (CNORM == 0.0_RK) THEN
    STOP 'SUBR. BAND_SHIFT: C IS THE ZERO MATRIX.'
  ENDIF
  !-----------------------------
#undef A
#undef B
#undef C
END SUBROUTINE BAND_SHIFT
!==============================================================================
!
!   BUILDS THE COMPLEX SHIFTED MATRIX "C := A - MU * IDENT", WHERE BOTH 
!   "A" AND "B" ARE BANDED REAL SYMMETRIC MATRICES, "MU" IS THE COMPLEX
!   SHIFT PARAMETER, AND "C" IS COMPLEX SYMMETRIC.
!
!==============================================================================
SUBROUTINE BAND_SHIFT_C_STD (BAND_A, N, NLD, NA, MU, BAND_C, NC, CNORM)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),   INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX (NOT CHANGED).
INTEGER,    INTENT(IN):: N            ! SIZE OF MATRIX "A" AND "C".
INTEGER,    INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "A" AND "C".
INTEGER,    INTENT(IN):: NA           ! LEADING DIMENSION OF "A".
COMPLEX(RK),INTENT(IN):: MU            ! SHIFT PARAMETER IN COMPLEX NUMBER.
COMPLEX(RK),INTENT(OUT):: BAND_C(NC,N) ! SHIFTED MATRIX "C" ("C:=A-MU*IDENT").
INTEGER,    INTENT(IN):: NC            ! LEADING DIMENSION OF "BAND_C".
REAL(RK),   INTENT(OUT):: CNORM        ! NORM OF MATRIX "C".
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: STDERR = 0 ! LOGICAL UNIT NUMBER OF STDERR FOR UNIX.
COMPLEX(RK) W
REAL(RK) S
INTEGER I, J, M
100 FORMAT('SUBR. BAND_SHIFT_C_STD WAS CALLED WITH', 1X, A, I5)
200 FORMAT('SUBR. BAND_SHIFT_C_STD WAS CALLED WITH', 1X, A, I6, A, I6)
#define A(I,J) BAND_A(M-(I)+(J),(I))
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-----------------------------------------------------------------------
  ! CHECK ARGUMENTS.
  IF (N <= 2) THEN
    WRITE(STDERR,100) 'N SMALLER THAN 3: N=', N
    STOP 'ERROR: BAND_SHIFT_C_STD.'
  ELSEIF (NA < NLD + 1) THEN
    WRITE(STDERR,200) 'NA=', NA, ' < NLD+1=', NLD+1
    STOP 'ERROR: BAND_SHIFT_C_STD.'
  ELSEIF (NLD >= N) THEN
    WRITE(STDERR,200) 'NLD=', NLD, ' >= N=', N
    STOP 'ERROR: BAND_SHIFT_C_STD.'
  ELSEIF (NC <= 3 * NLD) THEN
    WRITE(STDERR,200) 'NC=', NC, ' <= 3*NLD=', 3*NLD
    STOP 'ERROR: BAND_SHIFT_C_STD.'
  ENDIF
  !---------------------
  M = NLD + 1
  CNORM = 0.0_RK
  DO  I = 1, N
    S = 0.0_RK
    DO  J = MAX(1,I-NLD), I
      IF (I == J) THEN
        W = A(I,J) - MU
      ELSE
        W = A(I,J)
      ENDIF
      S = S + ABS(W)
      C(I,J) = W
      C(J,I) = W
    ENDDO
    DO  J = I+1, MIN(I+NLD,N)
      S = S + ABS(A(J,I))
    ENDDO
    DO  J = (I+NLD)+1, MIN(I+2*NLD,N)
      C(I,J) = 0.0_RK
    ENDDO
    CNORM = MAX(CNORM,S)
  ENDDO
  IF (CNORM == 0.0_RK) THEN
    STOP 'SUBR. BAND_SHIFT_C_STD: MATRIX C IS A ZERO MATRIX.'
  ENDIF
  !-----------------------------------------------------
#undef A
#undef C
END SUBROUTINE BAND_SHIFT_C_STD
!==============================================================================
!
!  BUILDS THE REAL SHIFTED MATRIX "C := A - MU * IDENT", WHERE BOTH
!  "A" AND "B" ARE BANDED REAL SYMMETRIC MATRICES, "MU" IS THE REAL SHIFT 
!  PARAMETER, AND THE MATRIX "C" IS REAL SYMMETRIC.
!
!==============================================================================
SUBROUTINE BAND_SHIFT_STD (BAND_A, N, NLD, NA, MU, BAND_C, NC, CNORM)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN) :: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A"(NOT CHANGED).
INTEGER, INTENT(IN) :: N            ! SIZE OF MATRIX "A".
INTEGER, INTENT(IN) :: NLD          ! LOWER BANDWIDTH OF "A".
INTEGER, INTENT(IN) :: NA           ! LEADING DIMENSION OF "A".
REAL(RK),INTENT(IN) :: MU           ! SHIFT PARAMETER IN REAL NUMBER.
REAL(RK),INTENT(OUT):: BAND_C(NC,N) ! SHIFTED MATRIX "C" ("C:=A-MU*IDENT").
INTEGER, INTENT(IN) :: NC           ! LEADING DIMENSION OF "BAND_C".
REAL(RK),INTENT(OUT):: CNORM        ! NORM OF MATRIX "C".
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: STDERR = 0  ! LOGICAL UNIT NUMBER FOR STDERR FOR UNIX.
REAL(RK) T, S
INTEGER I, J, K, M, KMX
100 FORMAT('SUBR. BAND_SHIFT_STD WAS CALLED WITH', 1X, A, I5)
200 FORMAT('SUBR. BAND_SHIFT_STD WAS CALLED WITH', 1X, A, I6, A, I6)
#define A(I,J) BAND_A(M-(I)+(J),(I))
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-----------------------------------------------------------------------
  ! CHECK ARGUMENTS.
  IF (N <= 2) THEN
    WRITE(STDERR, 100) 'N SMALLER THAN 3: N=', N
    STOP 'ERROR: BAND_SHIFT_STD.'
  ELSEIF (NA < NLD + 1) THEN
    WRITE(STDERR, 200) 'NA=', NA, ' < NLD+1=', NLD+1
    STOP 'ERROR: BAND_SHIFT_STD.'
  ELSEIF (NLD >= N) THEN
    WRITE(STDERR, 200) 'NLD=', NLD, ' >= N=', N
    STOP 'ERROR: BAND_SHIFT_STD.'
  ELSEIF (NC <= 3 * NLD) THEN
    WRITE(STDERR, 200) 'NC=', NC, ' <= 3*NLD=', 3*NLD
    STOP 'ERROR: BAND_SHIFT_STD.'
  ENDIF
  !---------------------
  M = NLD + 1
  CNORM = 0.0_RK
  DO  I = 1, N
    S = 0.0_RK
    DO  J = MAX(1,I-NLD), I
      IF (J == I) THEN
        T = A(I,J) - MU
      ELSE
        T = A(I,J)
      ENDIF
      C(I,J) = T
      C(J,I) = T
      S = S + ABS(T)
    ENDDO
    DO  J = I+1, MIN(I+NLD,N)
      S = S + ABS(A(J,I))
    ENDDO
    DO  J = (I+NLD)+1, MIN(I+2*NLD,N)
      C(I,J) = 0.0_RK
    ENDDO
    CNORM = MAX(CNORM,S)
  ENDDO
  IF (CNORM == 0.0_RK) THEN
    STOP 'SUBR. BAND_SHIFT_STD: MATRIX C IS A ZERO MATRIX.'
  ENDIF
  !-----------------------------
#undef A
#undef C
END SUBROUTINE BAND_SHIFT_STD
!******************************************************************************
!*
!* NOTE: THE ROUTINES: 
!*    "BAND_DECOMP_C", "BAND_LU_SOLVE_C", "BAND_LU_SOLVE_C_NV"
!*    "BAND_DECOMP", "BAND_LU_SOLVE" ARE DERIVED FROM MURATA'S ROUTINE 
!*    "INVITR", WHICH IS FOR THE REAL SYMMETRIC BANDED MATRIX.
!*  REF: MURATA, K. AND OGUNI, T. AND KARAKI, Y.:
!*      "SUPERCOMPUTERS AND THEIR APPLICATIONS
!*      FOR SCIENTIFIC & ENGINEERING COMPUTATIONS",
!*      (WRITTEN IN JAPANESE), MARUZEN, TOKYO, 1985.
!*
!******************************************************************************
!==============================================================================
!
!   LU-DECOMPOSITION OF THE MATRIX "C" WITH PARTIAL (ROW) PIVOTING,
!   WHERE THE MATRIX IS BANDED COMPLEX-SYMMETRIC.
!
!==============================================================================
SUBROUTINE BAND_DECOMP_C (N, NLD, EPS, CNORM, BAND_C, NC, IPIV)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER,    INTENT(IN)   :: N            ! SIZE OF MATRIX "C".
INTEGER,    INTENT(IN)   :: NLD          ! LOWER BANDWIDTH OF "C".
REAL(RK),   INTENT(IN)   :: EPS          ! TOLERANCE OF DECOMPOSITION.
REAL(RK),   INTENT(IN)   :: CNORM        ! NORM OF MATRIX "C".
COMPLEX(RK),INTENT(INOUT):: BAND_C(NC,N) ! MATRIX "C" AND ITS "LU"-FACTORS.
INTEGER,    INTENT(IN)   :: NC           ! LEADING DIMENSION OF "BAND_C".
INTEGER,    INTENT(OUT)  :: IPIV(N)      ! PIVOT PERMUTATION DATA.
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: STDERR = 0 ! LOGICAL UNIT NUMBER OF STDERR FOR UNIX.
INTEGER,PARAMETER:: FILE_99 = 99
REAL(RK) UEPS, EPS1, EPS2, EPS3
COMPLEX(RK) CMX, W
REAL(RK) S
INTEGER I, J, K, M, KMX
INTEGER NS_CNT, IE, JE
100 FORMAT('SUBR. BAND_DECOMP_C WAS CALLED WITH', 1X, A, I5)
200 FORMAT('SUBR. BAND_DECOMP_C WAS CALLED WITH', 1X, A, I6, A, I6)
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-----------------------------------------------------------------------
  M = NLD + 1
  !-- CHECK ARGUMENTS.
  IF (N <= 2) THEN
    WRITE(STDERR, 100) 'N SMALLER THAN 3: N=', N
    STOP 'ERROR: BAND_DECOMP_C.'
  ELSEIF (NLD >= N) THEN
    WRITE(STDERR, 200) 'NLD=', NLD, ' >= N=', N
    STOP 'ERROR: BAND_DECOMP_C.'
  ELSEIF (NC <= 3 * NLD) THEN
    WRITE(STDERR, 200) 'NC=', NC, ' <= 3*NLD=', 3*NLD
    STOP 'ERROR: BAND_DECOMP_C.'
  ENDIF
  !---------------------
  IF (CNORM == 0.0_RK) THEN
    STOP 'SUBR. BAND_DECOMP_C: MATRIX C IS A ZERO MATRIX.'
  ENDIF
  !-----------------------------
  UEPS = MAX(EPS, EPSILON(EPS))
  EPS1 = MAX(CNORM * UEPS, EPS)
  EPS2 = CNORM * 1.E-3
  EPS3 = N * EPS1
  !-----------------------------------------------------
  !  TRIANGULAR DECOMPOSITION WITH INTERCHANGES.
  NS_CNT = 0
  DO  K = 1, N
    IE = MIN(K+NLD,N)
    JE = MIN(K+2*NLD,N)
    !---
    CMX = C(K,K)
    KMX = K
    DO  I = K+1, IE
      IF (ABS(C(I,K)) <= ABS(CMX)) CYCLE
      CMX = C(I,K)
      KMX = I
    ENDDO
    !---
    IPIV(K) = KMX
    CMX = C(KMX,K)
    IF (ABS(CMX) == 0.0_RK) CMX = EPS1
    IF (ABS(CMX) <= EPS3) THEN ! SMALL PIVOT.
      NS_CNT = NS_CNT + 1  ! COUNT THE NUMBER OF SMALL PIVOTS.
    ENDIF
    !---
    IF (KMX /= K) THEN ! EXCHANGE ROWS.
      DO  J = K, JE
        W = C(KMX,J)
        C(KMX,J) = C(K,J)
        C(K,J) = W
      ENDDO
    ENDIF
    !
    C(K,K) = CMX
    DO  I = K+1, IE
      C(I,K) = C(I,K) / CMX
      W = C(I,K)
      DO  J = K+1, JE
        C(I,J) = C(I,J) - W * C(K,J)
      ENDDO
    ENDDO
    !
  ENDDO
#ifdef DEBUG
  IF (NS_CNT > 0) THEN
    WRITE( *,     900) NS_CNT
    WRITE(FILE_99,900) NS_CNT
  ENDIF
#else
  IF (NS_CNT > 1) THEN
    WRITE( *,     900) NS_CNT
    WRITE(FILE_99,900) NS_CNT
  ENDIF
#endif
  900 FORMAT('SUBR. BAND_DECOMP_C: NS_CNT=', I6)
  !----
#undef C
END SUBROUTINE BAND_DECOMP_C
!==============================================================================
!
!  SOLVES A LINEAR EQUATION WHOSE COMPLEX-SYMMETRIC COEFFICIENT MATRIX "C"
!  HAS ALREADY BEEN LU-FACTORED. BOTH THE RIGHT-HAND-SIDE VECTOR AND 
!  THE SOLUTION VECTOR "X" ARE COMPLEX.
!
!==============================================================================
SUBROUTINE BAND_LU_SOLVE_C (N, NLD, X, BAND_C, NC, IPIV)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER,    INTENT(IN):: N            ! SIZE OF MATRIX "C".
INTEGER,    INTENT(IN):: NLD          ! LOWER BANDWIDTH OF "C".
COMPLEX(RK),INTENT(INOUT):: X(N)      ! R-H-S AND SOLUTION VECTOR.
COMPLEX(RK),INTENT(IN):: BAND_C(NC,N) ! PRE-STORED "LU"-FACTORS OF MATRIX "C".
INTEGER,    INTENT(IN):: NC           ! LEADING DIMENSION OF "BAND_C".
INTEGER,    INTENT(IN):: IPIV(N)      ! PRE-STORED PIVOT PERMUTATION DATA.
!-----------------------------------------------------------------------
COMPLEX(RK) W
INTEGER I, J, K, KMX, M
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-----------------------------------------------------------------------
  M = NLD + 1
  !-----------------------------------------------------
  ! FORWARD ELIMINATION PROCESS.
  DO  K = 1, N-1
    KMX = IPIV(K)
    W = X(KMX)
    X(KMX) = X(K)
    X(K) = W
    DO  I = K+1, MIN(K+NLD,N)
      X(I) = X(I) - W * C(I,K)
    ENDDO
  ENDDO
  !-----------------------------------------------------
  ! BACKWARD SUBSTITUTION PROCESS.
  DO  K = N, 1, -1
    W = 0.0_RK
    DO  J = K+1, MIN(K+2*NLD,N)
      W = W + C(K,J) * X(J)
    ENDDO
    X(K) = (X(K) - W) / C(K,K)
  ENDDO
  !-----------------------------------------------------
#undef C
END SUBROUTINE BAND_LU_SOLVE_C
!==============================================================================
!    
!   SOLVES "NV" LINEAR EQUATIONS WHOSE COEFFICIENTS MATRICES ARE THE SAME
!   COMPLEX-SYMMETRIC MATRIX "C", WHICH HAS ALREADY BEEN LU-FACTORED.
!   BOTH THE SET OF RIGHT-HAND-SIDE VECTORS AND THE SET OF SOLUTION VECTORS
!   "X" ARE COMPLEX.
!
!==============================================================================
SUBROUTINE BAND_LU_SOLVE_C_NV (N, NLD, NV, X, BAND_C, NC, IPIV)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER,    INTENT(IN)   :: N           ! SIZE OF MATRIX "C".
INTEGER,    INTENT(IN)   :: NLD         ! LOWER BANDWIDTH OF "C".
INTEGER,    INTENT(IN)   :: NV          ! NUMBER OF SIMULTANEOUS VECTORS.
COMPLEX(RK),INTENT(INOUT):: X(N,NV)     ! SET OF VECTORS FOR RHS AND SOLUTION.
COMPLEX(RK),INTENT(IN)   :: BAND_C(NC,N)! PRE-STORED LU-FACTORS OF MATRIX "C".
INTEGER,    INTENT(IN)   :: NC          ! LEADING DIMENSION OF "BAND_C".
INTEGER,    INTENT(IN)   :: IPIV(N)     ! PRE-STORED PIVOT PERMUTATION DATA.
!-----------------------------------------------------------------------
COMPLEX(RK) W(NV)
INTEGER I, J, K, M, KMX
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-----------------------------------------------------------------------
  M = NLD + 1
  !-----------------------------------------------------
  ! FORWARD ELIMINATION PROCESS.
  DO  K = 1, N-1
    KMX = IPIV(K)
    W(:) = X(KMX,:)
    X(KMX,:) = X(K,:)
    X(K,:) = W(:)
    DO  I = K+1, MIN(K+NLD,N)
      X(I,:) = X(I,:) - W(:) * C(I,K)
    ENDDO
  ENDDO
  !-----------------------------------------------------
  ! BACKWARD SUBSTITUTION PROCESS.
  DO  K = N, 1, -1
    W(:) = 0.0_RK
    DO  J = K+1, MIN(K+2*NLD,N)
      W(:) = W(:) + C(K,J) * X(J,:)
    ENDDO
    X(K,:) = (X(K,:) - W(:)) / C(K,K)
  ENDDO
  !-----------------------------------------------------
#undef C
END SUBROUTINE BAND_LU_SOLVE_C_NV
!==============================================================================
!
!   THE LU-DECOMPOSITION OF THE MATRIX "C" WITH THE PARTIAL (ROW) PIVOTING,
!   WHERE THE MATRIX IS REAL COMPLEX-SYMMETRIC.
!
!==============================================================================
SUBROUTINE BAND_DECOMP (N, NLD, EPS, CNORM, BAND_C, NC, IPIV)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN)   :: N            ! SIZE OF MATRIX "C".
INTEGER, INTENT(IN)   :: NLD          ! LOWER BANDWIDTH OF "C".
REAL(RK),INTENT(IN)   :: EPS          ! TOLERANCE OF DECOMPOSITION.
REAL(RK),INTENT(IN)   :: CNORM        ! NORM OF MATRIX "C".
REAL(RK),INTENT(INOUT):: BAND_C(NC,N) ! MATRIX "C" AND ITS "LU"-FACTORS.
INTEGER, INTENT(IN)   :: NC           ! LEADING DIMENSION OF "BAND_C".
INTEGER, INTENT(OUT)  :: IPIV(N)      ! PIVOT PERMUTATION DATA.
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: STDERR = 0 ! LOGICAL UNIT NUMBER FOR STDERR FOR UNIX.
INTEGER,PARAMETER:: FILE_99 = 99
REAL(RK) UEPS, EPS1, EPS2, EPS3
REAL(RK) CMX, T
INTEGER I, J, K, M, KMX
INTEGER NS_CNT, IE, JE
100 FORMAT('SUBR. BAND_DECOMP WAS CALLED WITH', 1X, A, I5)
200 FORMAT('SUBR. BAND_DECOMP WAS CALLED WITH', 1X, A, I6, A, I6)
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-----------------------------------------------------------------------
  M = NLD + 1
  !-- CHECK ARGUMENTS.
  IF (N <= 2) THEN
    WRITE(STDERR, 100) 'N SMALLER THAN 3: N=', N
    STOP 'ERROR: BAND_DECOMP.'
  ELSEIF (NLD >= N) THEN
    WRITE(STDERR, 200) 'NLD=', NLD, ' >= N=', N
    STOP 'ERROR: BAND_DECOMP.'
  ELSEIF (NC <= 3 * NLD) THEN
    WRITE(STDERR, 200) 'NC=', NC, ' <= 3*NLD=', 3*NLD
    STOP 'ERROR: BAND_DECOMP.'
  ENDIF
  !---------------------
  IF (CNORM == 0.0_RK) THEN
    STOP 'SUBR. BAND_DECOMP: THE MATRIX C IS A ZERO MATRIX.'
  ENDIF
  !-----------------------------
  UEPS = MAX(EPS, EPSILON(EPS))
  EPS1 = MAX(CNORM * UEPS, EPS)
  EPS2 = CNORM * 1.E-3
  EPS3 = N * EPS1
  !-----------------------------------------------------
  !  TRIANGULAR DECOMPOSITION WITH ROW INTERCHANGES.
  NS_CNT = 0
  DO  K = 1, N
    IE = MIN(K+NLD, N)
    JE = MIN(K+2*NLD, N)
    !---
    CMX = C(K,K)
    KMX = K
    DO  I = K+1, IE
      IF (ABS(C(I,K)) <= ABS(CMX)) CYCLE
      CMX = C(I,K)
      KMX = I
    ENDDO
    !---
    IPIV(K) = KMX
    CMX = C(KMX,K)
    IF (CMX == 0.0_RK) CMX = EPS1
    !
    IF (ABS(CMX) <= EPS3) THEN  ! SMALL PIVOT.
      NS_CNT = NS_CNT + 1  ! COUNT THE NUMBER OF SMALL PIVOTS.
    ENDIF
    !---
    IF (KMX /= K) THEN ! EXCHANGE ROWS.
      DO  J = K, JE
        T = C(KMX,J)
        C(KMX,J) = C(K,J)
        C(K,J) = T
      ENDDO
    ENDIF
    !
    C(K,K) = CMX
    DO  I = K+1, IE
      C(I,K) = C(I,K) / CMX
      T = C(I,K)
      DO  J = K+1, JE
        C(I,J) = C(I,J) - T * C(K,J)
      ENDDO
    ENDDO
    !
  ENDDO
#ifdef DEBUG
   IF (NS_CNT > 0) THEN
     WRITE( *,     900) NS_CNT
     WRITE(FILE_99,900) NS_CNT
   ENDIF
#else
   IF (NS_CNT > 1) THEN
     WRITE( *,     900) NS_CNT
     WRITE(FILE_99,900) NS_CNT
   ENDIF
#endif
   900 FORMAT('SUBR. BAND_DECOMP: NS_CNT=', I6)
   !----
#undef C
END SUBROUTINE BAND_DECOMP
!==============================================================================
!
!  SOLVE A LINEAR EQUATION OF A REAL BANDED MATRIX "C" USING THE 
!  PREVIOUS "LU"-FACTORS.
!
!==============================================================================
SUBROUTINE BAND_LU_SOLVE (N, NLD, X, BAND_C, NC, IPIV)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN) :: N            ! SIZE OF MATRIX "C".
INTEGER, INTENT(IN) :: NLD          ! LOWER BANDWIDTH OF "C".
REAL(RK),INTENT(INOUT):: X(N)       ! R-H-S AND SOLUTION VECTOR.
REAL(RK),INTENT(IN) :: BAND_C(NC,N) ! PRE-STORED "LU"-FACTORS.
INTEGER, INTENT(IN) :: NC           ! LEADING DIMENSION OF "BAND_C".
INTEGER, INTENT(IN) :: IPIV(N)      ! PRE-STORED PERMUTATION DATA.
!-----------------------------------------------------------------------
REAL(RK) T
INTEGER I, J, K, KMX, M
#define C(I,J) BAND_C(M-(I)+(J),(I))
!-------------------------------------
  M = NLD + 1
  !==============================
  !  FORWARD ELIMINATION PROCESS.
  DO  K = 1, N-1
    KMX = IPIV(K)
    T = X(KMX)
    X(KMX) = X(K)
    X(K) = T
    DO  I = K+1, MIN(K+NLD,N)
      X(I) = X(I) - T * C(I,K)
    ENDDO
  ENDDO
  !==============================
  ! BACKWARD SUBSTITUTION PROCESS.
  DO  K = N, 1, -1
    T = 0.0_RK
    DO  J = K+1, MIN(K+2*NLD,N)
      T = T + C(K,J) * X(J)
    ENDDO
    X(K) = (X(K) - T) / C(K,K)
  ENDDO
  !==============================
#undef C
END SUBROUTINE BAND_LU_SOLVE
