!******************************************************************************
!
!$Id: hqr_orth.F90,v 1.1 2010/07/10 01:44:28 murakami Exp $
!
!******************************************************************************
!==============================================================================
!
!  HOUSEHOLDER  QR-FACTORIZATION AS  "A P = Q R" ("P" IS THE PERMUTATION).
!  THE EFFECTIVE RANK OF MATRIX "A" IS DETERMINED.
!  NOTE: WITH THIS CODE, "Q" IS OVER-WRITTEN TO THE PLACE OF "A".
!
!==============================================================================
SUBROUTINE HQR_ORTH (M, N, A, LDA, R, LDR, RANK, PERMU, REL_THRES)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN)   :: M         ! NUMBER OF ROWS IN "A"
INTEGER, INTENT(IN)   :: N         ! NUMBER OF COLUMNS IN "A"
REAL(RK),INTENT(INOUT):: A(LDA,N)  ! MATRIX "A(M,N)".
INTEGER, INTENT(IN)   :: LDA       ! LEADING DIMENSION OF "A".
REAL(RK),INTENT(OUT)  :: R(LDR,N)  ! UPPER TRIANGULAR MATRIX "R(N,N)".
INTEGER, INTENT(IN)   :: LDR       ! LEADING DIMENSION OF "R".
INTEGER, INTENT(OUT)  :: RANK      ! EFFECTIVE RANK DETERMINED BY HQR.
INTEGER, INTENT(OUT)  :: PERMU(N)  ! COLUMN PIVOT PERMUTATION DATA.
REAL(RK),INTENT(IN)   :: REL_THRES ! RELATIVE CUT-OFF THRESHOLD.
!-----------------------------------------------------------------------
REAL(RK) T, ALPHA, SGM, PIV(N)
REAL(RK) ANRM, ALEN2(N), ALEN
INTEGER  I, J, K, MX_IND, I_T
!------------------------------
  IF (M < N) THEN
    STOP 'SUBR. HQR_ORTH: ERROR M < N.'
  ENDIF
  DO  J = 1, N
    ALEN2(J) = SUM(A(:M,J)**2)
    PERMU(J) = J
  ENDDO
  !--------------------------------
  RANK = 0
  LOOP_K: DO  K = 1, N
    !--
    MX_IND = MAXLOC(ALEN2(K:N), DIM=1) + (K-1)
    IF (MX_IND > N) STOP 'SUBR. HQR_ORTH: ERROR, MX_IND > N'
    !--
    ALEN = SQRT(ALEN2(MX_IND))
    !--
    IF (K == 1) THEN
      ANRM = ALEN
      IF (ANRM == 0.0_RK) EXIT LOOP_K
    ELSE
      IF (ALEN <= ANRM * REL_THRES) EXIT LOOP_K
    ENDIF
    RANK = K
    IF (MX_IND /= K) THEN
      I_T = PERMU(K)
      PERMU(K) = PERMU(MX_IND)
      PERMU(MX_IND) = I_T
      DO  I = 1, M  !  IN CASE "R" IS NOT REQUIRED, THIS LINE IS "DO I=K,M"
        T = A(I,K)
        A(I,K) = A(I,MX_IND)
        A(I,MX_IND) = T
      ENDDO
    ENDIF
    !=======================
    ! HOUSEHOLDER TRANSFORMATION.
    SGM = SIGN(ALEN, A(K,K))
    PIV(K) = SGM * (SGM + A(K,K))
    A(K,K) = A(K,K) + SGM
    ALPHA = - 1.0_RK / PIV(K)
    !-------------------
    R(:K-1,K) = A(:K-1,K)
    R(K,K) = - SGM
    R(K+1:N,K) = 0.0_RK ! ELEMENTS BELOW THE DIAGONAL ARE SET TO ZEROS.
    !-------------------
    DO  J = K+1, N
      T = DOT_PRODUCT(A(K:M,K), A(K:M,J)) * ALPHA
      !-------
      I = K
      A(I,J) = A(I,J) + A(I,K) * T
      ALEN2(J) = 0.0_RK
      DO  I = K+1, M
        A(I,J) = A(I,J) + A(I,K) * T
        ALEN2(J) = ALEN2(J) + A(I,J)**2
      ENDDO
    ENDDO
  ENDDO LOOP_K
  !=======================
  ! STORE R(:,RANK+1:N) FROM "A".
  DO  J = RANK+1,N
    R(:RANK,J) = A(:RANK,J)
    R(RANK+1:N,J) = 0.0_RK ! ROWS BELOW RANK ARE SET TO ZEROS.
  ENDDO
  !=======================
  ! HOUSEHOLDER BACKWARD-TRANSFORMATIONS TO FORM "Q" EXPLICITLY.
  DO  K = RANK, 1, -1
    ALPHA = - 1.0_RK / PIV(K)
    DO  J = RANK, K+1, -1
      T  = DOT_PRODUCT(A(K:M,K), A(K:M,J)) * ALPHA
      A(K:M,J) = A(K:M,J) + A(K:M,K) * T
    ENDDO
    A(:K-1,K) = 0.0_RK
    T = A(K,K) * ALPHA
    A(K,K) = 1.0_RK + A(K,K) * T
    A(K+1:M,K) =  + A(K+1:M,K) * T
  ENDDO
  !===
#ifdef DEBUG
  PRINT *, 'SUBR. HQR_ORTH: THE 2-NORM OF VECTORS "Q" ARE:'
  DO  J = 1, RANK
    PRINT *, J, SQRT(SUM(A(:M,J)**2))
  ENDDO
#endif
END SUBROUTINE HQR_ORTH
