!******************************************************************************
!
! $Id: main_gen_mp.F90,v 1.3 2010/07/12 12:36:15 murakami Exp $
!
!******************************************************************************
!==============================================================================
!
! FOR THE GENERALIZED EIGENPROBLEM.
!
!  A TEST PROGRAM OF THE FILTER DIAGONALIZATION METHOD FOR THE BANDED
!  SYMMETRIC DEFINITE GENERALIZED EIGENPROBLEM "Ax=\lambda Bx".
!  THE FILTER IS A LINEAR COMBINATION OF RESOLVENTS.
!
!==============================================================================
PROGRAM MAIN_GEN_MP
!$ USE OMP_LIB
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER  NPROB        ! PROBLEM NUMBER.
INTEGER  N            ! SIZE OF BOTH MATRICES "A" AND "B".
INTEGER  NLD          ! LOWER BANDWIDTH OF BOTH MATRICES "A" AND "B".
REAL(RK) LAM0, LAM1   ! BOTH ENDS OF THE INTERVAL.
REAL(RK) MU           ! SHARPNESS OF FILTER ATTENUATION. 
REAL(RK) A_MAX, A_MIN ! BOTH UPPER AND LOWER BOUNDS OF FILTER ATTENUATION. 
INTEGER  FILT_TYPE    ! TYPE OF FILTER. 
INTEGER  NF           ! (HALF OF) THE DEGREE OF THE FILTER FUNCTION.
INTEGER  NV           ! NUMBER OF RANDOM VECTORS.
REAL(RK) REL_THRES    ! RELATIVE THRESHOLD FOR SVD FOR CUT-OFF.
INTEGER  MX_ITER      ! THE MAX NUMBER OF INVERSE ITERATIONS.
REAL(RK),ALLOCATABLE:: BAND_A(:,:) ! BANDED SYMMETRIC MATRIX "A".
REAL(RK),ALLOCATABLE:: BAND_B(:,:) ! BANDED SYMMETRIC MATRIX "B".
REAL(RK),ALLOCATABLE:: X(:,:) ! SET OF INPUT VECTORS FOR THE FILTER.
REAL(RK),ALLOCATABLE:: Y(:,:) ! SET OF VECTORS.
COMPLEX(RK),ALLOCATABLE:: ZPOLE(:) ! COMPLEX POLES IN NORMAL COORDINATE.
COMPLEX(RK),ALLOCATABLE:: ZCOEF(:) ! COMPLEX COEFFICIENTS FOR POLES.
REAL(RK)               :: CINF     ! COEFFICIENT FOR THE CONSTANT TERM(UNUSED).
REAL(RK),ALLOCATABLE:: SIGMA(:) ! SINGULAR VALUES.
REAL(RK),ALLOCATABLE:: Q(:)     ! APPROXIMATED EIGENVALUES.
REAL(RK),ALLOCATABLE:: MAGIN(:), MAGOUT(:) ! RATES OF ROUND-OFF MAGNIFICATIONS.
INTEGER IFLAG ! STATUS FLAG FOR ALLOCATION.
!------
INTEGER,PARAMETER:: FILE_66 = 66
INTEGER,PARAMETER:: FILE_77 = 77
INTEGER RANK_Y      ! THE EFFECTIVE RANK OF "Y" DETERMINED.
INTEGER L
REAL*8  TM(0:6)     ! ARRAY OF THE WALL-CLOCK TIMES.
CHARACTER(80) FMT200, FMT300
CHARACTER(30) STR_PREC
INTEGER NUM_FOUND
REAL(RK) EPS_A
!------------------------------------------------------------------------------
!$OMP PARALLEL
!$OMP SINGLE
!$  WRITE(*,      *) 'OMP_NUM_THREADS=', OMP_GET_NUM_THREADS()
!$  WRITE(FILE_66,*) 'OMP_NUM_THREADS=', OMP_GET_NUM_THREADS()
!$OMP END SINGLE
!$OMP END PARALLEL
  !------
  WRITE(*,      *) 'PROG. MAIN_GEN (GENERALIZED EIGENPROBLEM).'
  WRITE(FILE_66,*) 'PROG. MAIN_GEN (GENERALIZED EIGENPROBLEM).'
  !------
  CALL COMPARE_RK ! CHECK IF KIND-VALUE "RK" IS SAME TO THAT OF FILTER LIBRARY.
  !------
  CALL SET_STR_PREC ! SET THE STRING "STR_PREC".
  CALL SET_FMT200   ! SET THE FORMAT STRING "FMT200".
  WRITE(*,      100) RK, STR_PREC
  WRITE(FILE_66,100) RK, STR_PREC
  100 FORMAT('RK=', I2, 1X, A)
  !====================
  CALL SET_JOB_PARAMS (NPROB, N, NLD, LAM0, LAM1, &
                       FILT_TYPE, MU, A_MAX, A_MIN, NF, &
                       NV, REL_THRES, &
                       MX_ITER)
  !====================
  ! ALLOCATE ARRAYS FOR THE FILTER PARAMETERS.
  ALLOCATE (ZPOLE(NF), ZCOEF(NF), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'PROG. MAIN_GEN: FAILED TO ALLOCATE ZPOLE, ZCOEF.'
  !------
  ! CALCULATE THE NORMALIZED POLES AND THEIR COEFFICIENTS OF THE FILTER.
  EPS_A = SQRT(A_MAX-1)
  CALL FILTER_POLES_COEFS (FILT_TYPE, NF, MU, EPS_A, ZPOLE, ZCOEF, CINF)
  !====================
  ! ALLOCATE THE BAND MATRICES.
  ALLOCATE (BAND_A(NLD+1,N), BAND_B(NLD+1,N), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'PROG. MAIN_GEN: FAILED TO ALLOCATE BAND_A, BAND_B.'
  ! ALLOCATE THE RANDOM VECTORS.
  ALLOCATE (X(N,NV), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'PROG. MAIN_GEN: FAILED TO ALLOCATE X.'
  ! ALLOCATE THE FILTERED VECTORS.
  ALLOCATE (Y(N,NV), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'PROG. MAIN_GEN: FAILED TO ALLOCATE Y.'
  ! ALLOCATE AN ARRAY OF ROUND-OFF ERROR MAGNIFICATION RATES.
  ALLOCATE (MAGIN(NV), MAGOUT(NV), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'PROG. MAIN_GEN: FAILED TO ALLOCATE MAGIN, MAGOUT.'
  !====================
  ! SET THE VALUES OF THE BAND MATRICES A AND B.
  CALL SET_BAND_AB (NPROB, N, NLD, BAND_A, BAND_B)
  !------
  CALL MY_CLOCK (TM(0))
  !====================
  ! GENERATE "X", THE RANDOM INITIAL VECTORS.
  ! AVOID TO PARALLELIZE THIS TO PROVIDE THE INDEPENDENCE OF THE RANDOMNESS.
  DO  L = 1, NV
    CALL RANDOM_VECTOR (N, X(1,L))
  ENDDO
  !------
  CALL MY_CLOCK (TM(1))
  !====================
  ! B-ORTHONORMALIZE THE VECTORS IN "X".
  CALL B_ORTH_MGS (BAND_B, N, NLD, SIZE(BAND_B,1), NV, X)
  PRINT *, 'X IS B-ORTHONORMALIZED.'
  !------
  CALL MY_CLOCK (TM(2))
  !====================
  ! Y <-- FILTER (X).
  CALL APPLY_FILTER_NV (NF, ZPOLE, ZCOEF, CINF, &
                        LAM0, LAM1, &
                        BAND_A, BAND_B, N, NLD, &
                        SIZE(BAND_A,1), SIZE(BAND_B,1), &
                        X, Y, NV)
  DEALLOCATE (X)
  !------
  CALL MY_CLOCK (TM(3))
  !====================
  ! ORTHONORMALIZATION OF "Y" BY SVD.
  ALLOCATE (SIGMA(NV), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'PROG. MAIN_GEN: FAILED TO ALLOCATE SIGMA.'
  CALL B_ORTH_SVD (BAND_B, N, NLD, SIZE(BAND_B,1), &
                   NV, Y, REL_THRES, SIGMA, RANK_Y)
  MAGIN(:RANK_Y) = 1 / SIGMA(:RANK_Y)
  !------
  CALL MY_CLOCK (TM(4))
  !====================
  ! SUBSPACE METHOD APPLIED TO THE SET OF BASIS VECTORS.
  ALLOCATE (Q(RANK_Y), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'PROG. MAIN_GEN: FAILED TO ALLOCATE Q.'
  !------
  CALL RITZ (BAND_A, N, NLD, SIZE(BAND_A,1), &
             RANK_Y, Y, Q, LAM0, LAM1, MAGIN, MAGOUT)
  !------
  CALL MY_CLOCK (TM(5))
  !====================
  CALL G_INV_ITER_STEP (BAND_A, BAND_B, N, NLD, &
                        SIZE(BAND_A,1), SIZE(BAND_B,1), &
                        RANK_Y, Y, Q, MX_ITER)
  !------
  CALL MY_CLOCK (TM(6))
  !====================
  WRITE(      *, *) 'AFTER THE INVERSE ITERATIONS.'
  WRITE(FILE_66, *) 'AFTER THE INVERSE ITERATIONS.'
  WRITE(      *, *) '    THE VALUE OF Q;   IS IT IN THE INTERVAL?'
  WRITE(FILE_66, *) '    THE VALUE OF Q;   IS IT IN THE INTERVAL?'
  DO  L = 1, RANK_Y
    WRITE(      *, FMT200) L, Q(L), (LAM0<=Q(L).AND.Q(L)<=LAM1)
    WRITE(FILE_66, FMT200) L, Q(L), (LAM0<=Q(L).AND.Q(L)<=LAM1)
  ENDDO
  !------
  NUM_FOUND = COUNT(LAM0<=Q(:).AND.Q(:)<=LAM1)
  SELECT CASE(NUM_FOUND)
    CASE(0:1)   ; FMT300 ="('## FOUND ', I5, ' SOLUTION.')"
    CASE DEFAULT; FMT300 ="('## FOUND ', I5, ' SOLUTIONS.')"
  END SELECT
  WRITE(      *, FMT300) NUM_FOUND
  WRITE(FILE_66, FMT300) NUM_FOUND
  !------
  WRITE(      *, *) 'PROG. MAIN_GEN: NORMAL END.'
  WRITE(FILE_66, *) 'PROG. MAIN_GEN: NORMAL END.'
  !====================
  ! PRINT THE ELAPSED TIMES.
  CALL PRINTF_TIME (FILE_77, "THE TOTAL FDM PROCESS      ", TM(5)-TM(0))
  !------
  CALL PRINTF_TIME (FILE_77, "  GENERATE RANDOM VECTORS  ", TM(1)-TM(0))
  CALL PRINTF_TIME (FILE_77, "  B-ORTHONORMALIZE VECTORS ", TM(2)-TM(1))
  CALL PRINTF_TIME (FILE_77, "  FILTER INPUT VECTORS     ", TM(3)-TM(2))
  CALL PRINTF_TIME (FILE_77, "  B-SVD OUTPUT VEC ANALYSIS", TM(4)-TM(3))
  CALL PRINTF_TIME (FILE_77, "  CALCULATE RITZ VALUES    ", TM(5)-TM(4))
  CALL PRINTF_TIME (FILE_77, "INVERSE-ITERATIONS         ", TM(6)-TM(5))
  !------
  STOP
CONTAINS
  !---------------------------
  SUBROUTINE SET_STR_PREC
    SELECT CASE(RK)
      CASE(KIND(1.E0)); STR_PREC = '(SINGLE PRECISION)'
      CASE(KIND(1.D0)); STR_PREC = '(DOUBLE PRECISION)'
#ifndef NO_QUADRUPLE_PRECISION
      CASE(KIND(1.Q0)); STR_PREC = '(QUADRUPLE PRECISION)'
#endif
    END SELECT
  END SUBROUTINE
  !---------------------------
  SUBROUTINE SET_FMT200
    SELECT CASE(RK)
      CASE(KIND(1.E0)); FMT200 = "('Q(',I4,')=',ES14.7, ' : ',L)"
      CASE(KIND(1.D0)); FMT200 = "('Q(',I4,')=',ES22.15,' : ',L)"
#ifndef NO_QUADRUPLE_PRECISION
      CASE(KIND(1.Q0)); FMT200 = "('Q(',I4,')=',ES41.34,' : ',L)"
#endif
      CASE DEFAULT    ; FMT200 = "('Q(',I4,')=',ES22.15,' : ',L)"
    END SELECT
  END SUBROUTINE
  !---------------------------
END PROGRAM MAIN_GEN_MP
!==============================================================================
!
!   APPLY THE RESOLVENT-TYPE FILTER TO SIMULTANEOUSLY THE SET OF "NV" VECTORS.
!
!==============================================================================
SUBROUTINE APPLY_FILTER_NV (NF, ZPOLE, ZCOEF, CINF, &
                            LAM0, LAM1, &
                            BAND_A, BAND_B, N, NLD, NA, NB, X, Y, NV)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER,    INTENT(IN):: NF       ! NUMBER OF RESOLVENTS FOR THE FILTER.
COMPLEX(RK),INTENT(IN):: ZPOLE(NF)! NORMALIZED POLES OF THE FILTER.
COMPLEX(RK),INTENT(IN):: ZCOEF(NF)! NORMALIZED COEFFICIENTS OF THE FILTER.
REAL(RK),   INTENT(IN):: CINF     ! COEFFICIENT FOR THE IDENTITY(NOT IN USE).
REAL(RK),INTENT(IN):: LAM0, LAM1  ! INTERVAL OF THE PASSBAND OF THE FILTER.
REAL(RK),INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A".
REAL(RK),INTENT(IN):: BAND_B(NB,N) ! BANDED SYMMETRIC MATRIX "B".
INTEGER, INTENT(IN):: N            ! SIZE OF MATRICES OF THE EIGENPROBLEM.
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF BOTH "A" AND "B".
INTEGER, INTENT(IN):: NA, NB       ! LEADING DIMENSION OF "BAND_A", "BAND_B".
REAL(RK),INTENT(IN) :: X(N,NV) ! SET OF INPUT VECTORS TO THE FILTER.
REAL(RK),INTENT(OUT):: Y(N,NV) ! SET OF OUTPUT VECTORS FROM THE FILTER.
INTEGER, INTENT(IN):: NV       ! NUMBER OF VECTORS TO APPLY THE FILTER.
!------------------------------------------------------------------------------
REAL(RK),PARAMETER:: EPS_LU = EPSILON(1.0_RK) * 10
REAL(RK),   ALLOCATABLE:: BX(:,:)! STORE FOR "NV" REAL VECTORS.'
COMPLEX(RK),ALLOCATABLE:: Z(:,:) ! STORE FOR "NV" COMPLEX VECTORS.
COMPLEX(RK),ALLOCATABLE:: BAND_C(:,:) ! BANDED MATRIX "C" AND ITS "LU"-FACTORS.
INTEGER    ,ALLOCATABLE:: IPIV(:) ! STORE FOR PIVOT PERMUTATION DATA.
REAL(RK),   ALLOCATABLE:: YPART(:,:) ! THREAD LOCAL STORE FOR "NV VECTOR "Y".
INTEGER IFLAG ! STATUS FLAG FOR ALLOCATION.
COMPLEX(RK) ZRHO(NF), ZGAMMA(NF) ! SHIFTS AND COEFFICIENTS FOR RESOLVENT.
REAL(RK) CNORM
INTEGER I
!--------------------------------------------------
  ! THE SHIFTS OF THE RESOLVENTS AND THEIR COEFFICIENT OF LINEAR COMBINATION.
  ! "CINF" IS NOT TRANSFORMED AS "ZGAMMA" IS (BUT IS LEFT UNUSED AND OMITTED).
  ZRHO(:) = ZPOLE(:) * ((LAM1-LAM0)/2) + ((LAM0+LAM1)/2)
  ZGAMMA(:) = ZCOEF(:) * ((LAM1-LAM0)/2)
!$OMP  PARALLEL PRIVATE(BAND_C,Z,IPIV,BX,YPART,IFLAG,CNORM)
  !------
!$OMP  WORKSHARE
  Y(:N,:NV) = 0.0_RK ! ACCUMULATION IS MADE TO THIS VECTOR.
!$OMP END WORKSHARE
  !====================
  ALLOCATE (BX(N,NV), Z(N,NV), STAT=IFLAG)
   IF (IFLAG/=0) STOP 'SUBR.APPLY_FILTER_NV:FAILED TO ALLOCATE BX AND Z.'
  ALLOCATE (BAND_C(3*NLD+1,N), IPIV(N), STAT=IFLAG)
   IF (IFLAG/=0) STOP 'SUBR.APPLY_FILTER_NV:FAILED TO ALLOCATE BAND_C OR IPIV.'
  ALLOCATE (YPART(N,NV), STAT=IFLAG)
   IF (IFLAG/=0) STOP 'SUBR.APPLY_FILTER_NV:FAILED TO ALLOCATE YPART.'
  !====================
  YPART(:N,:NV) = 0.0_RK
!$OMP  DO
  DO  I = 1, NF ! THE COMPLEX CONJUGATE SYMMETRY IS USED.
#if 1
    WRITE(*,100) I
    100 FORMAT('DEBUG: (APPLY_FILTER_NV) I=', I5)
#endif
    !====================
    ! MAKE THE BAND MATRIX "C = A - ZRHO(I)*B",
    CALL BAND_SHIFT_C (BAND_A, BAND_B, N, NLD, &
                       SIZE(BAND_A,1), SIZE(BAND_B,1), &
                       ZRHO(I), BAND_C, SIZE(BAND_C,1), CNORM)
    !====================
    ! AND DECOMPOSE THE BAND MATRIX "C" TO "LU"-FACTORS.
    CALL BAND_DECOMP_C (N, NLD, EPS_LU, CNORM, BAND_C, SIZE(BAND_C,1), IPIV)
    !====================
    ! B * X -> Z.
    CALL BAND_MULT_NV (BAND_B, N, NLD, SIZE(BAND_B,1), NV, X, BX)
    !====================
    ! SIMULTANEOUSLY SOLVE THE LINEAR EQUATIONS: (A - ZRHO(I)*B) Z = BX
    ! THE PREVIOUS LU-FACTORS ARE REUSED.
    Z(:N,:NV) = BX(:N,:NV)
    CALL BAND_LU_SOLVE_C_NV (N, NLD, NV, Z, BAND_C, SIZE(BAND_C,1), IPIV)
    !====================
    ! ADD TO YPART THE REAL PARTS OF PRODUCTS OF "ZGAMMA(I)" AND VECTOR "Z".
    ! THE FACTOR TWO COMES FROM THE COMPLEX CONJUGATE PAIR.
    YPART(:N,:NV) = YPART(:N,:NV) + &
        (REAL((2*ZGAMMA(I)),KIND=RK) * REAL(Z(:N,:NV),KIND=RK) &
        -AIMAG((2*ZGAMMA(I)))        * AIMAG(Z(:N,:NV)))
    !====================
  ENDDO
!$OMP  END DO NOWAIT
  DEALLOCATE(BX, Z, BAND_C, IPIV)
!$OMP  CRITICAL
  Y(:N,:NV) = Y(:N,:NV) + YPART(:N,:NV) ! PARTIAL SUMS OF Y ARE SUMMED UP.
!$OMP  END CRITICAL
  DEALLOCATE (YPART)
!$OMP  END PARALLEL
END SUBROUTINE APPLY_FILTER_NV
!==============================================================================
!
!    INVERSE ITERATION PROCESS FOR THE APPRIXIMATED "NV" EIGENPAIRS
!    OF THE GENERALIZED EIGENPROBLEM.
!
!==============================================================================
SUBROUTINE G_INV_ITER_STEP (BAND_A, BAND_B, N, NLD, NA, NB, &
                            NV, Y, E, MX_ITER)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A".
REAL(RK),INTENT(IN):: BAND_B(NB,N) ! BANDED SYMMETRIC MATRIX "B".
INTEGER, INTENT(IN):: N          ! SIZE OF MATRICES OF THE EIGENPROBLEM.
INTEGER, INTENT(IN):: NLD        ! LOWER BANDWIDTH OF BOTH "A" AND "B".
INTEGER, INTENT(IN):: NA, NB     ! LEADING DIMENSIONS FOR "BAND_A" AND "BAND_B".
INTEGER, INTENT(IN):: NV         ! NUMBER OF APPROXIMATED EIGENPAIRS.
REAL(RK),INTENT(INOUT):: Y(N,NV) ! SET OF APPROXIMATED EIGENVECTORS.
REAL(RK),INTENT(INOUT):: E(NV)   ! APPROXIMATED EIGENVALUES.
INTEGER, INTENT(IN):: MX_ITER    ! MAX NUMBER OF INVERSE ITERATION.
!------------------------------------------------------------------------------
REAL(RK),PARAMETER:: EPS_LU = EPSILON(1.0_RK) * 10
INTEGER,PARAMETER:: FILE_99 = 99
REAL(RK),ALLOCATABLE:: BAND_C(:,:), LU_B(:,:) ! STORE OF "LU"-FACTORS.
INTEGER ,ALLOCATABLE:: IPIV_C(:), IPIV_B(:)   ! STORE OF PIVOT PERMUTATION DATA.
INTEGER IFLAG ! STATUS FLAG FOR THE ALLOCATION.
REAL(RK) LU_BNORM
INTEGER L
CHARACTER(40) FMT
INTEGER IT
REAL(RK),DIMENSION(0:MX_ITER,NV):: Q, DEL, MAG
!------------------------------------------------------------------------------
  WRITE(      *, *) 'INVERSE ITERATIONS FOR GENERALIZED EIGENPROBLEM.'
  WRITE(FILE_99, *) 'INVERSE ITERATIONS FOR GENERALIZED EIGENPROBLEM.'
  !====================
  ALLOCATE (LU_B(3*NLD+1,N), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. G_INV_ITER_STEP: FAILED TO ALLOCATE LU_B.'
  ALLOCATE (IPIV_B(N), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. G_INV_ITER_STEP: FAILED TO ALLOCATE IPIV_B.'
  !====================
  ! MAKE "LU" FACTOR OF "B" (THIS PART IS UGLY).
  CALL BAND_SHIFT (BAND_B, BAND_A, N, NLD, SIZE(BAND_B,1), SIZE(BAND_A,1), &
                   0.0_RK, LU_B, SIZE(LU_B,1), LU_BNORM)
  CALL BAND_DECOMP (N, NLD, EPS_LU, LU_BNORM, LU_B, SIZE(LU_B,1), IPIV_B)
  !====================
  ! FOR EACH APPROXIMATED EIGENPAIR, THE INVERSE ITERATION IS APPLIED.
!$OMP PARALLEL PRIVATE(BAND_C,IPIV_C,IFLAG)
  ALLOCATE (BAND_C(3*NLD+1,N), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. G_INV_ITER_STEP: FAILED TO ALLOCATE BAND_C.'
  ALLOCATE (IPIV_C(N), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. G_INV_ITER_STEP: FAILED TO ALLOCATE IPIV_C.'
!$OMP DO
  DO  L = 1, NV
    CALL G_INV_ITR (BAND_A, BAND_B, N, NLD, SIZE(BAND_A,1), SIZE(BAND_B,1), &
                    Y(1,L), BAND_C, SIZE(BAND_C,1), IPIV_C, &
                    LU_B, SIZE(LU_B,1), IPIV_B, &
                    MX_ITER, Q(0,L), DEL(0,L), MAG(0,L))
  ENDDO
!$OMP ENDDO
  DEALLOCATE (BAND_C, IPIV_C)
!$OMP END PARALLEL
  DEALLOCATE (LU_B, IPIV_B)
  !====================
  SELECT CASE (RK)
    CASE (KIND(1.E0)); FMT="(1X, 'Q=', ES14.7)"
    CASE (KIND(1.D0)); FMT="(1X, 'Q=', ES22.15)"
#ifndef NO_QUADRUPLE_PRECISION
    CASE (KIND(1.Q0)); FMT="(1X, 'Q=', ES41.34)"
#endif
    CASE DEFAULT     ; FMT="(1X, 'Q=', ES22.15)"
  END SELECT
  100 FORMAT('L=', I5)
  200 FORMAT(1X, 'IT', I1, ':')
  300 FORMAT(1X, 'DEL=', ES9.1 : 1X, 'MAG=', ES9.1)
  !----
  DO  L = 1, NV
    WRITE(      *, 100) L
    WRITE(FILE_99, 100) L
    LOOP: DO  IT = 0, MX_ITER
      WRITE(      *, 200, ADVANCE='NO') IT
      WRITE(FILE_99, 200, ADVANCE='NO') IT
      WRITE(      *, FMT, ADVANCE='NO') Q(IT,L)
      WRITE(FILE_99, FMT, ADVANCE='NO') Q(IT,L)
      IF (IT == 0) THEN
        WRITE(      *, 300) DEL(IT,L)
        WRITE(FILE_99, 300) DEL(IT,L)
      ELSE
        WRITE(      *, 300) DEL(IT,L), MAG(IT,L)
        WRITE(FILE_99, 300) DEL(IT,L), MAG(IT,L)
      ENDIF
    ENDDO LOOP
    E(L) = Q(MX_ITER,L)
  ENDDO
END SUBROUTINE G_INV_ITER_STEP
!==============================================================================
!
!  RAYLEIGH INVERSE ITERATION PROCESS FOR AN APPROXIMATED EIGENPAIR 
!  OF THE GENERALIZED EIGENPROBLEM.
!
!==============================================================================
SUBROUTINE G_INV_ITR (BAND_A, BAND_B, N, NLD, NA, NB, &
                      V, BAND_C, NC, IPIV_C, &
                      LU_B, NLU_B, IPIV_B, MX_ITER, Q, DEL, MAG)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_A(NA,N)! BANDED SYMMETRIC MATRIX "A".
REAL(RK),INTENT(IN):: BAND_B(NB,N)! BANDED SYMMETRIC MATRIX "B".
INTEGER, INTENT(IN):: NA, NB      ! LEADING DIMENSIONS OF "BAND_A" AND "BAND_B".
INTEGER, INTENT(IN):: N           ! SIZE OF MATRICES "A" AND "B".
INTEGER, INTENT(IN):: NLD         ! LOWER BANDWIDTH OF BOTH "A" AND "B".
REAL(RK),INTENT(INOUT):: V(N) ! APPROXIMATED EIGENVECTOR.
REAL(RK)           :: BAND_C(NC,N) ! WORK AREA FOR "LU"-FACTORS.
INTEGER, INTENT(IN):: NC           ! LEADING DIMENSION OF "BAND_C".
INTEGER            :: IPIV_C(N)    ! WORK AREA FOR PIVOT PERMUTATION DATA.
REAL(RK),INTENT(IN):: LU_B(NLU_B,N) ! PREVIOUSLY-STORED "LU"-FACTORS OF "B".
INTEGER, INTENT(IN):: NLU_B         ! LEADING DIMENSION OF "LU_B".
INTEGER, INTENT(IN):: IPIV_B(N)     ! PREVIOUSLY-STORED PIVOT FOR "LU_B".
INTEGER, INTENT(IN):: MX_ITER ! MAX NUMBER OF ITERATIONS.
REAL(RK),INTENT(OUT),DIMENSION(0:MX_ITER):: Q, DEL, MAG 
                    ! ARRAY OF EIGENVALUES, DELS, MAGS FOR STEPS OF ITERATION.
!------------------------------------------------------------------------------
REAL(RK),PARAMETER:: EPS_LU = EPSILON(1.0_RK) * 10
REAL(RK),ALLOCATABLE:: R(:), S(:)
INTEGER IFLAG ! STATUS FLAG FOR ALLOCATION.
REAL(RK) CNORM
INTEGER ITER
!-----------------------------------
  ALLOCATE (R(N), S(N), STAT=IFLAG)
  IF (IFLAG/=0) STOP 'SUBR. G_INV_ITR: FAILED TO ALLOCATE R, S.'
  !------------
  LOOP: DO  ITER = 0, MX_ITER
    IF (ITER > 0) THEN
      !====================
      ! COMPUTE "V := B * V".
      S(:N) = V(:N)
      CALL BAND_MULT (BAND_B, N, NLD, SIZE(BAND_B,1), S, V) 
      !====================
      ! SOLVE THE LINEAR EQUATION: "(A - Q * B) V_NEXT = V".
      CALL BAND_SHIFT (BAND_A, BAND_B, N, NLD, &
                       SIZE(BAND_A,1), SIZE(BAND_B,1), &
                       Q(ITER-1), BAND_C, SIZE(BAND_C,1), CNORM)
      CALL BAND_DECOMP (N, NLD, EPS_LU, CNORM, BAND_C, SIZE(BAND_C,1), IPIV_C)
      CALL BAND_LU_SOLVE (N, NLD, V, BAND_C, NC, IPIV_C)
      !====================
    ENDIF
    !====================
    ! COMPUTE "S := B * V".
    CALL BAND_MULT (BAND_B, N, NLD, SIZE(BAND_B,1), V, S) 
    !====================
    ! "MAG(ITER)" IS SET TO "SQRT((V, B*V))", THE "B"-NORM OF "V".
    MAG(ITER) = SQRT(DOT_PRODUCT(V(:N), S(:N)))
    !====================
    ! "V" IS "B"-NORMALIZED, "S" IS ALSO SCALED BY THE SAME FACTOR.
    V(:N) = V(:N) / MAG(ITER) 
    S(:N) = S(:N) / MAG(ITER)
    !====================
    ! COMPUTE "R := A * V".
    CALL BAND_MULT (BAND_A, N, NLD, SIZE(BAND_A,1), V, R) 
    !====================
    ! UPDATE THE RAYLEIGH QUOTIENT "Q := (V, AV)".
    Q(ITER) = DOT_PRODUCT(V(:N), R(:N)) 
    !====================
    ! COMPUTE THE RESIDUAL: "R := (A-Q*B)*V = (A*V)-Q*(B*V)".
    R(:N) = R(:N) - Q(ITER) * S(:N) 
    !====================
    ! COMPUTE "S := B^{-1} * R".
    S(:N) = R(:N)
    CALL BAND_LU_SOLVE (N, NLD, S, LU_B, SIZE(LU_B,1), IPIV_B)
    !====================
    ! COMPUTE "DEL:=SQRT((R,B^{-1}*R))", AN ERROR BOUND OF THE EIGENVALUE.
    DEL(ITER) = SQRT(DOT_PRODUCT(R(:N), S(:N)))
    !====================
  ENDDO LOOP
  DEALLOCATE (R, S)
  !----
END SUBROUTINE G_INV_ITR
