!******************************************************************************
!
! $Id: misc.F90,v 1.2 2010/07/12 12:36:15 murakami Exp $
!
!******************************************************************************
!* THIS ROUTINE IS MOSTLY EQUIVALENT TO THE ALGOL-60 ROUTINE "JACOBI"
!* IN THE BOOK:
!*     J.H. WILKINSON AND C. REINSCH, "LINEAR ALGEBRA ,
!*     VOL.II OF HANDBOOK FOR AUTOMATIC COMPUTATION",
!*     SPRINGER-VERLAG, NEW YORK, PP.202-211, 1971.
!* OR THE ORIGINAL PAPER :
!*     H. RUTISHAUSER, "THE JACOBI METHOD FOR REAL SYMMETRIC MATRICES",
!*     NUMER. MATH. VOL.9, PP.1-10, 1966.
!*
!* B.T. SMITH ET AL.,: "MATRIX EIGENSYSTEM ROUTINES -- EISPACK GUIDE",
!* 2ND ED., VOL.6 OF LECTURE NOTES IN COMPUTER SCIENCE,
!* SPRINGER-VERLAG, NEW YORK, 1976.
!*
!******************************************************************************
!==============================================================================
!
!  THE THRESHOLD JACOBI'S METHOD.
!  ONLY THE DIAGONALS AND RIGHT UPPER HALF OF THE MATRIX "A" IS TO BE SUPPLIED.
!  SUPER-DIAGONAL ELEMENTS OF THE MATRIX "A" ARE DESTROYED.
!  THE DIAGONAL AND SUB-DIAGONALS OF THE MATRIX "A" ARE UNCHANGED.
!  IN THE MATRIX NOTATION: "D = U^{T} A U",  OR  "A = U D U^{T}".
!
!==============================================================================
SUBROUTINE RSYM_JACOBI (N, A, NA, D, U, NU, NROT)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN)   :: N      ! SIZE OF MATRIX "A".
INTEGER, INTENT(IN)   :: NA     ! LEADING DIMENSION OF "A".
REAL(RK),INTENT(INOUT):: A(NA,N)! SYMMETRIC MATRIX "A" (WILL BE DESTROYED).
REAL(RK),INTENT(OUT):: D(N)   ! EIGENVALUES CALCULATED.
INTEGER, INTENT(IN) :: NU     ! LEADING DIMENSION OF "U".
REAL(RK),INTENT(OUT):: U(NU,N)! ORTHOGONAL MATRIX "U".
INTEGER, INTENT(OUT):: NROT ! NUMBER OF ROTATIONS PERFORMED FOR INFORMATIVE.
!-----------------------------------------------------------------------
CHARACTER(*),PARAMETER:: PROGNAME='RSYM_JACOBI'
INTEGER,PARAMETER:: MAX_ITR=50
REAL(RK) B(N), Z(N)
REAL(RK) SM, TRESH, G, H, TAU, THETA, C, S, T
INTEGER IP, IQ, J, ITER
!-----------------------------------------------------------------------
  U(:N,:N) = 0.0_RK
  DO  IP = 1, N
    U(IP,IP) = 1.0_RK
    B(IP) = A(IP,IP)
  ENDDO
  D = B
  Z = 0.0_RK
  !----------------
  NROT = 0
  LOOP: DO  ITER = 1, MAX_ITR
    SM = 0.0_RK
    DO  IQ = 2, N
    DO  IP = 1, IQ-1
      SM = SM + ABS(A(IP,IQ))
    ENDDO
    ENDDO
    IF (SM == 0.0_RK) EXIT LOOP ! THIS IS THE NORMAL EXIT.
    IF (ITER < 4) THEN
      TRESH = 0.2 * SM / N**2
    ELSE
      TRESH = 0.0_RK
    ENDIF
    DO  IQ = 2, N
    DO  IP = 1, IQ-1
      G = 100.0 * ABS(A(IP,IQ))
      IF ((ITER > 4) .AND. (ABS(D(IP))+G == ABS(D(IP)) ) &
                     .AND. (ABS(D(IQ))+G == ABS(D(IQ)) )) THEN
        A(IP,IQ) = 0.0_RK
      ELSEIF (ABS(A(IP,IQ)) > TRESH) THEN
        H = D(IQ) - D(IP)
        IF (ABS(H)+G == ABS(H)) THEN
          T = A(IP,IQ) / H
        ELSE
          THETA = 0.5_RK * H / A(IP,IQ)
          T = 1.0_RK / (ABS(THETA)+SQRT(1.0_RK+THETA**2))
          IF (THETA < 0.0_RK) T = -T
        ENDIF
        C = 1.0_RK / SQRT(1.0_RK+T**2)
        S = T * C
        TAU = S / (1.0_RK+C)
        H = T * A(IP,IQ)
        Z(IP) = Z(IP) - H
        Z(IQ) = Z(IQ) + H
        D(IP) = D(IP) - H
        D(IQ) = D(IQ) + H
        A(IP,IQ) = 0.0_RK
        DO  J = 1, IP-1
          G = A(J,IP)
          H = A(J,IQ)
          A(J,IP) = G - S * (H + G * TAU)
          A(J,IQ) = H + S * (G - H * TAU)
        ENDDO
        DO  J = IP+1, IQ-1
          G = A(IP,J)
          H = A(J,IQ)
          A(IP,J) = G - S * (H + G * TAU)
          A(J,IQ) = H + S * (G - H * TAU)
        ENDDO
        DO  J = IQ+1, N
          G = A(IP,J)
          H = A(IQ,J)
          A(IP,J) = G - S * (H + G * TAU)
          A(IQ,J) = H + S * (G - H * TAU)
        ENDDO
        DO  J = 1, N
          G = U(J,IP)
          H = U(J,IQ)
          U(J,IP) = G - S * (H + G * TAU)
          U(J,IQ) = H + S * (G - H * TAU)
        ENDDO
        NROT = NROT + 1
      ENDIF
    ENDDO
    ENDDO
    B = B + Z
    D = B
    Z = 0.0_RK
  ENDDO LOOP
  IF (ITER > MAX_ITR) THEN
    WRITE( *,*) PROGNAME, ': ERROR.', MAX_ITR, 'ITERATIONS EXCEEDS THE LIMIT.'
    WRITE(66,*) PROGNAME, ': ERROR.', MAX_ITR, 'ITERATIONS EXCEEDS THE LIMIT.'
    STOP
  ENDIF
END SUBROUTINE RSYM_JACOBI
!==============================================================================
!
!   SORT THE ALL EIGENPAIRS IN ORDER WITH RESPECT TO THEIR EIGENVALUES.
!   (THE CODING STYLE IS UGLY, BUT IT WORKS.)
!
!==============================================================================
SUBROUTINE EIG_SORT (N, D, U, NU, W, NW, DIRECTION)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN)   :: N       ! NUMBER OF EIGENPAIRS.
REAL(RK),INTENT(INOUT):: D(N)    ! EIGENVALUES TO BE SORTED.
REAL(RK),INTENT(INOUT):: U(NU,N) ! VECTORS TO BE SORTED AS "D".
INTEGER, INTENT(IN)   :: NU      ! LEADING DIMENSION OF "U".
REAL(RK)              :: W(NW,N) ! WORK AREA.
INTEGER, INTENT(IN)   :: NW      ! LEADING DIMENSION OF "W".
CHARACTER(*),INTENT(IN):: DIRECTION ! STRING 'ASCENDING' OR 'DESCENDING'.
!-----------------------------------------------------------------------
INTEGER,EXTERNAL:: COMPAR2_ASCEND, COMPAR2_DESCEND
! NOTE, FOR IEEE-754 FLOAT SYSTEM (4BYTES, 8BYTES, 16BYTES FOR SP, DP, QP).
INTEGER,PARAMETER:: BYTES_PER_SP=4, BYTES_PER_DP=8, BYTES_PER_QP=16
INTEGER BYTES_PER_WORD
REAL(RK) D2(2,N)
INTEGER IND(N), J
!-----------------------------------------------------------------------
  IF (NU < N .OR. NW < N) THEN
    PRINT *, 'N=', N, ',NU=', NU, ',NW=', NW
    STOP 'SUBR. EIG_SORT: ERROR NU < N .OR NW < N.'
  ENDIF
  DO  J = 1, N
    D2(1,J) = D(J)
    D2(2,J) = REAL(J, KIND=RK)
  ENDDO
  !---
  SELECT CASE (RK)
    CASE (KIND(1.E0)); BYTES_PER_WORD = BYTES_PER_SP ! SINGLE PRECISION
    CASE (KIND(1.D0)); BYTES_PER_WORD = BYTES_PER_DP ! DOUBLE PRECISION
#ifndef NO_QUADRUPLE_PRECISION
    CASE (KIND(1.Q0)); BYTES_PER_WORD = BYTES_PER_QP ! QUADRUPLE PRECISION.
#endif
    CASE DEFAULT; STOP 'SUBR. EIG_SORT: UNKNOWN REAL-KIND'
  END SELECT
  !---
  SELECT CASE (DIRECTION(1:1))
    CASE ('A','a'); CALL MY_QSORT (D2, N, 2*BYTES_PER_WORD, COMPAR2_ASCEND)
    CASE ('D','d'); CALL MY_QSORT (D2, N, 2*BYTES_PER_WORD, COMPAR2_DESCEND)
    CASE DEFAULT; STOP 'SUBR. EIG_SORT: UNKNOWN DIRECTION.'
  END SELECT
  !----
  IND(:N) = NINT(D2(2,:N))
  D(:N) = D2(1,:N)
  !----
  W(:N,:N) = U(:N,:N)
  U(:N,:N) = W(:N,IND(:N))
  !----
END SUBROUTINE EIG_SORT
!==============================================================================
!
!
!==============================================================================
FUNCTION COMPAR2_ASCEND (X, Y) RESULT(VAL)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER  VAL
REAL(RK),INTENT(IN):: X(2), Y(2)
!-----------------------------------------------------------------------
  IF (X(1) < Y(1)) THEN
    VAL = -1
  ELSEIF (X(1) > Y(1)) THEN
    VAL = 1
  ELSE
    VAL = 0
  ENDIF
END FUNCTION COMPAR2_ASCEND
!==============================================================================
!
!
!==============================================================================
FUNCTION COMPAR2_DESCEND (X, Y) RESULT(VAL)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER          :: VAL
REAL(RK),INTENT(IN):: X(2), Y(2)
!-----------------------------------------------------------------------
  IF (X(1) > Y(1)) THEN
    VAL = -1
  ELSEIF (X(1) < Y(1)) THEN
    VAL = 1
  ELSE
    VAL = 0
  ENDIF
END FUNCTION COMPAR2_DESCEND
!==============================================================================
!
!   MY_QSORT : QUICK SORT.
!   FORTRAN90 IMPLEMENTATION OF HOAR'S QUICK SORT.
!
!==============================================================================
SUBROUTINE MY_QSORT (A, N, SIZE, COMPAR)
IMPLICIT NONE
INTEGER,INTENT(IN)     :: SIZE      ! SIZE OF AN ITEM.
INTEGER,INTENT(IN)     :: N         ! NUMBER OF ITEMS.
CHARACTER,INTENT(INOUT):: A(SIZE,N) ! DATA AREA TO SORT.
INTEGER,EXTERNAL       :: COMPAR    ! EXTERNAL FUNCTION FOR COMPARISON.
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: MAXSP = 64 ! SUFFICIENT FOR THE 64-BIT ADDRESS SPACE.
INTEGER STACK(2,0:MAXSP-1)
INTEGER SP
CHARACTER X(SIZE)
INTEGER I, J, LEFT, RIGHT
!-----------------------------------------------------------------------
  CALL INIT_STACK
  CALL PUSH (1,N)
  DO WHILE (STACK_IS_NOT_EMPTY())
    CALL POP (LEFT,RIGHT)
    DO WHILE (LEFT < RIGHT)
      X(:) = A(:,(LEFT+RIGHT)/2)
      I = LEFT; J = RIGHT
      DO WHILE (I <= J)
        DO WHILE (COMPAR(A(1,I),X) < 0); I = I + 1; ENDDO
        DO WHILE (COMPAR(X,A(1,J)) < 0); J = J - 1; ENDDO
        IF (I <= J) THEN
          CALL SWAP (A(1,I), A(1,J))
          I = I + 1; J = J - 1
        ENDIF
      ENDDO
      !
      IF (J-LEFT < RIGHT-I) THEN
        IF (I < RIGHT) CALL PUSH (I, RIGHT)
        RIGHT = J
      ELSE
        IF (LEFT < J) CALL PUSH (LEFT, J)
        LEFT = I
      ENDIF
    ENDDO
  ENDDO
CONTAINS !----------------
  SUBROUTINE SWAP (STR1, STR2)
    CHARACTER,INTENT(INOUT):: STR1(SIZE), STR2(SIZE)
    CHARACTER TMP(SIZE)
    TMP = STR1; STR1 = STR2; STR2 = TMP
  END SUBROUTINE
  SUBROUTINE INIT_STACK
    SP = 0
  END SUBROUTINE
  FUNCTION STACK_IS_NOT_EMPTY () RESULT(VAL)
    LOGICAL VAL
    VAL = (SP > 0)
  END FUNCTION
  SUBROUTINE PUSH (P, Q)
    INTEGER,INTENT(IN):: P, Q
    STACK(1,SP) = P; STACK(2,SP) = Q; SP = SP + 1
  END SUBROUTINE
  SUBROUTINE POP (P, Q)
    INTEGER,INTENT(OUT):: P, Q
    SP = SP - 1; P = STACK(1,SP); Q = STACK(2,SP)
  END SUBROUTINE
END SUBROUTINE MY_QSORT

#undef USE_UNIX_ETIME
!==============================================================================
!
!    RETURNS THE WALL-CLOCK TIME IN SECONDS FROM SYSTEM CLOCK.
!
!==============================================================================
SUBROUTINE MY_CLOCK (SEC)
IMPLICIT NONE
REAL*8,INTENT(OUT):: SEC ! VALUE OF ELAPSED TIME IN SECOND.
!-----------------------------------------------------------------------
#ifdef USE_UNIX_ETIME
  REAL TARRAY(2)
  REAL*8,EXTERNAL:: ETIME ! SYSTEM DEPENDENT UNIX ROUTINE(NOT SO PORTABLE).
  SEC = ETIME(TARRAY)
  RETURN
#else
  INTRINSIC SYSTEM_CLOCK ! FORTRAN90 INTRINSIC ROUTINE.
  INTEGER CNT, CNT_RATE, CNT_MAX
  LOGICAL,SAVE:: FIRST_VISIT = .TRUE.
  INTEGER,SAVE:: NWRAP    ! THE NUMBER OF TIMER_WRAPS.
  INTEGER,SAVE:: LAST_CNT ! THE LAST COUNT
!-----------------------------------------------------------------------
  CALL SYSTEM_CLOCK (COUNT=CNT, COUNT_RATE=CNT_RATE, COUNT_MAX=CNT_MAX)
  IF (FIRST_VISIT) THEN
    FIRST_VISIT = .FALSE.
    NWRAP = 0
  ELSE
    IF (CNT < LAST_CNT) NWRAP = NWRAP + 1
  ENDIF
  SEC = (CNT + (1+DBLE(CNT_MAX)) * NWRAP) / CNT_RATE
  LAST_CNT = CNT
#endif
END SUBROUTINE MY_CLOCK
!==============================================================================
!
!
!==============================================================================
SUBROUTINE PRINTF_TIME (NFILE, MSG, T)
IMPLICIT NONE
INTEGER,     INTENT(IN):: NFILE ! LOGICAL UNIT NUMBER FOR THE FILE.
CHARACTER(*),INTENT(IN):: MSG   ! MESSAGE STRING.
REAL*8,      INTENT(IN):: T     ! THE TIME INTERVAL VALUE.
!-----------------------------------------------------------------------
  WRITE(*    , 100) MSG, T
  WRITE(NFILE, 100) MSG, T
  100 FORMAT(A32, 3X, F12.4)
END SUBROUTINE PRINTF_TIME

#ifdef USE_UNIX_ETIME
!==============================================================================
!
!   ELAPSE TIME IN LOW ACCURACY. (USUALLY 0.01 SECOND IN RESOLUTION)
!   ETIME IS THE UNIX SYSTEM DEPENDENT ROUTINE.
!
!==============================================================================
REAL*8 FUNCTION MY_SECOND ()
IMPLICIT NONE
REAL TARRAY(2)
REAL*8,EXTERNAL:: ETIME
  MY_SECOND = ETIME(TARRAY) ! THE SYSTEM DEPENDENT UNIX TIMER ROUTINE "ETIME".
END FUNCTION MY_SECOND
#endif
!==============================================================================
!
!    MAKE A RANDOM VECTOR.
!
!==============================================================================
SUBROUTINE RANDOM_VECTOR (N, V)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN) :: N    ! SIZE OF VECTOR "V".
REAL(RK),INTENT(OUT):: V(N) ! NORMALIZED VECTOR OF UNIFORM RANDOM NUMBER.
!-----------------------------------------------------------------------
REAL(RK) S
!----------------------------
  ! GENERATE THE UNIFORM RANDOM NUMBER IN THE UNIT INTERVAL (0,1).
  CALL RANDOM_NUMBER (V(:N))  
  ! CONVERT THE INTERVAL OF DISTRIBUTION TO (-1,1).
  V(:N) = 2 * V(:N) - 1.0_RK  
  ! THE VECTOR "V" IS NORMALIZED BY 2-NORM.
  S = SQRT(SUM(V(:N)**2))     
  V(:N) = V(:N) / S
END SUBROUTINE RANDOM_VECTOR
!==============================================================================
! THIS ROUTINE CHECKS IF THE KIND-VALUE USED IN THE FILTER LIBRARY 
! IS THE SAME TO THE ONE IN THIS ROUTINE, TO AVOIDS THE POSSIBLE MIXTURE.
!==============================================================================
SUBROUTINE COMPARE_RK
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
!---------------------------------
INTEGER,EXTERNAL:: RK_FILTER
!---------------------------------
  IF (RK /= RK_FILTER()) THEN
    WRITE (*, 100)
    WRITE (66,100)
    100 FORMAT('COMPARE_RK: KIND-VALUE OF THE FILTER LIBRARY IS DIFFERENT.')
    WRITE (*, 200)
    WRITE (66,200)
    200 FORMAT('PLEASE RECOMPILE THE FILTER LIBRARY WITH THE SAME KIND-VALUE.')
    STOP 'COMPARE_RK: DETECTED INCOMPATIBILITY OF THE KIND-VALUE.'
  ENDIF
END SUBROUTINE COMPARE_RK
