!******************************************************************************
!
! $Id: orth.F90,v 1.3 2010/07/12 12:36:15 murakami Exp $
!
!******************************************************************************
!==============================================================================
!
!   ORTHONORMALIZE NV VECTORS IN "X" BY THE MODIFIED GRAM-SCHMIDT METHOD.
!
!==============================================================================
SUBROUTINE ORTH_MGS (N, NV, X)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER,INTENT(IN)    :: N       ! SIZE OF VECTORS IN "X".
INTEGER,INTENT(IN)    :: NV      ! NUMBER OF VECTORS IN "X".
REAL(RK),INTENT(INOUT):: X(N,NV) ! SET OF VECTORS "X".
!------------------------------------------------------------------------------
REAL(RK) RLL, RLJ
INTEGER L, J
!----------------------------------------------------------
  DO  L = 1, NV
    RLL = SQRT(SUM(X(:,L)**2))
    X(:,L) = X(:,L) / RLL
    DO  J = L+1, NV
      RLJ = DOT_PRODUCT(X(:,L), X(:,J))
      X(:,J) = X(:,J) - RLJ * X(:,L)
    ENDDO
  ENDDO
END SUBROUTINE ORTH_MGS
!==============================================================================
!
!   ORTHONORMALIZATION BY THE CLASSICAL SVD METHOD (WITH REFINEMENT).
!
!==============================================================================
SUBROUTINE ORTH_SVD (N, NV, Y, REL_THRES, SIGMA, RANK)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN)   :: N       ! SIZE OF EACH VECTOR.
INTEGER, INTENT(IN)   :: NV      ! NUMBER OF VECTORS.
REAL(RK),INTENT(INOUT):: Y(N,NV) ! SET OF VECTORS TO ORTHONORMALIZE.
REAL(RK),INTENT(IN) :: REL_THRES ! RELATIVE THRESHOLD FOR SINGULAR VALUES.
REAL(RK),INTENT(OUT):: SIGMA(NV) ! SINGULAR VALUES.
INTEGER, INTENT(OUT):: RANK      ! EFFECTIVE RANK DETERMINED BY SVD.
!------------------------------------------------------------------------------
INTEGER,PARAMETER:: MX_LP=2 ! TWICE OF ITERATION SEEMS ENOUGH.
LOGICAL,PARAMETER:: SHOW_SVD_SIGMA=.TRUE.
INTEGER,PARAMETER:: FILE_66=66
INTEGER,PARAMETER:: FILE_88=88 ! DATA FILE FOR SINGULAR VALUES FOR GNUPLOT.
CHARACTER(*),PARAMETER:: FNAME_88 = 'sigma.data'
REAL(RK),ALLOCATABLE:: S(:,:), U(:,:), W(:,:)
INTEGER IFLAG ! STATUS FLAG OF ALLOCATION.
INTEGER LP, L, NROT
REAL(RK) ERR
!----------------------------------------------------------
  ALLOCATE (S(NV,NV), U(NV,NV), W(NV,NV), STAT=IFLAG)
  IF (IFLAG/=0) STOP 'ORHT_SVD: FAILED TO ALLOCATE S, U, W'
  LOOP: DO  LP = 1, MX_LP
    !====================
    ! CALCULATE THE SELF INNER PRODUCT: "S := Y^{T} * Y".
#if 0
    S(:NV,:NV) = MATMUL(TRANSPOSE(Y(:N,:NV)), Y(:N,:NV))
#else
    CALL MAT_SELF_DOT_PRODUCT (N, NV, Y, SIZE(Y,1), S, SIZE(S,1))
#endif
    !====================
    ! EIGENVALUE DECOMPOSITION OF "S" TO OBTAIN THE SQUARES
    ! OF THE SINGULAR VALUES "SIGMA" AND THE ROTATION MATRIX "U".
    CALL RSYM_JACOBI (NV, S, SIZE(S,1), SIGMA, U, SIZE(U,1), NROT)
#if 1
    PRINT 100, 'ORTH_SVD LP=', LP, ', RSYM_JACOBI NROT=', NROT
    100 FORMAT (A, I1, A, I10)
#endif
    ! SINGULAR PAIRS ARE SORTED IN DESCENDING ORDER BY THEIR SINGULAR VALUES.
    CALL EIG_SORT (NV, SIGMA, U, SIZE(U,1), W, SIZE(W,1), 'DESCENDING')
    !====================
    ! ROTATE "Y" BY THE ORTHOGONAL MATRIX "U" TO MAKE A SET OF ORTHOGONAL BASIS.
#if 0
    Y(:N,:NV) = MATMUL(Y(:N,:NV), U(:NV,:NV))
#else
    CALL RIGHT_TRFM (N, NV, Y, SIZE(Y,1), U, SIZE(U,1))
#endif
    !====================
  ENDDO LOOP
  !====================
  ! SINGULAR VALUES ARE CALCULATED FROM THEIR SQUARES.
  DO  L = 1, NV
    IF (SIGMA(L) > 0.0_RK) THEN
      SIGMA(L) = SQRT(SIGMA(L))
    ELSE
      SIGMA(L) = 0.0_RK
    ENDIF
  ENDDO
  !====================
  ! THE EFFECTIVE RANK IS DETERMINED BY THE RELATIVE THRESHOLD "REL_THRES".
  RANK = 0
  IF (SIGMA(1) > 0.0_RK) THEN
    DO  L = 1, NV
      IF (SIGMA(L) <= SIGMA(1) * REL_THRES) EXIT
      RANK = RANK + 1
    ENDDO
  ENDIF
  !====================
  ! NORMALIZE THOSE BASIS WHOSE SINGULAR VALUES ARE ABOVE THE THRESHOLD.
  DO  L = 1, RANK
    Y(:N,L) = Y(:N,L) / SIGMA(L)
  ENDDO
#if 1
  ! THIS ZERO-PADDING IS NOT NECESSARY.
  Y(:N,RANK+1:NV) = 0.0_RK 
#endif
  !====================
  ! SHOW THE CALCULATED SINGULAR VALUES AND THE EFFECTIVE RANK.
  IF (SHOW_SVD_SIGMA) THEN
    WRITE(      *,*) 'SINGULAR VALUES:'
    WRITE(FILE_66,*) 'SINGULAR VALUES:'
    DO  L = 1, NV
      WRITE(      *,200) L, SIGMA(L)
      WRITE(FILE_66,200) L, SIGMA(L)
      200 FORMAT(3X, 'SIGMA(', I3, ') =', ES9.1)
    ENDDO
    WRITE(      *,300) RANK
    WRITE(FILE_66,300) RANK
    300 FORMAT('RANK=', I3)
  ENDIF
  !====================
  ! WRITE FILE_88 THE DATA OF SINGULAR VALUES AND THE THRESHOLD FOR GNUPLOT.
  OPEN(FILE_88, FILE=FNAME_88, FORM='FORMATTED')
  DO  L = 1, NV
    WRITE(FILE_88, FMT=310) L, SIGMA(L)
    310 FORMAT(I3, 3X, ES9.1)
  ENDDO
  WRITE(FILE_88,'(//)')
  WRITE(FILE_88, FMT=310)  1, SIGMA(1)*REL_THRES
  WRITE(FILE_88, FMT=310) NV, SIGMA(1)*REL_THRES
  CLOSE(FILE_88)
  !====================
#if 1
  ! CHECK WHETHER THE SET OF SINGULAR VECTORS ARE ORTHONORMAL.
  ! (USUALLY, THIS CHECK IS NOT NECESSARY).
  CALL MAT_SELF_DOT_PRODUCT (N, RANK, Y, SIZE(Y,1), S, SIZE(S,1))
  ERR = 0.0_RK
  DO  L = 1, RANK
    IF (L > 1) ERR = MAX(ERR, MAXVAL(ABS(S(1:L-1,L))))
    ERR = MAX(ERR, ABS(S(L,L)-1.0_RK))
  ENDDO
  WRITE(      *,400) ERR 
  WRITE(FILE_66,400) ERR
  400 FORMAT('** MAX ORTHOGONAL ERROR OF SINGULAR VECTORS=', ES8.1)
#endif
  !====================
  DEALLOCATE (S, U, W)
END SUBROUTINE ORTH_SVD
!==============================================================================
!
!   HOUSEHOLDER-"QR" METHOD TO ORTHONORMALIZE THE SET OF VECTORS "Y".
!
!==============================================================================
SUBROUTINE ORTH_HQR (N, NV, Y, REL_THRES, R_DIAG, RANK_Y)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER,INTENT(IN)    :: N       ! SIZE OF EACH VECTOR.
INTEGER,INTENT(IN)    :: NV      ! NUMBER OF VECTORS.
REAL(RK),INTENT(INOUT):: Y(N,NV) ! SET OF VECTORS TO ORTHONORMALIZE.
REAL(RK),INTENT(IN) :: REL_THRES   ! RELATIVE THRESHOLD FOR PIVOT VALUES.
REAL(RK),INTENT(OUT):: R_DIAG(NV)  ! CALCULATED MAGNITUDES OF PIVOT VALUES.
INTEGER,INTENT(OUT) :: RANK_Y      ! EFFECTIVE RANK OF "Y" WITH THE THRESHOLD.
!------------------------------------------------------------------------------
LOGICAL,PARAMETER:: SHOW_HQR_PIVOT=.TRUE.
REAL(RK),PARAMETER:: EPS_HQR = EPSILON(1.0_RK)*10
INTEGER,PARAMETER:: FILE_66=66 !
REAL(RK),ALLOCATABLE:: R(:,:)   ! STORE OF RIGHT UPPER MATRIX.
INTEGER ,ALLOCATABLE:: PERMU(:) ! STORE OF PIVOT PERMUTATION DATA.
INTEGER IFLAG ! STATUS FLAG FOR ALLOCATION.
INTEGER L, RANK
!----------------------------------------------------------
  ALLOCATE (R(NV,NV), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. ORTH_HQR: FAILED TO ALLOCATE RDUM.'
  ALLOCATE (PERMU(NV),   STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. ORTH_HQR: FAILED TO ALLOCATE PERMU.'
  !====================
  ! HOUSEHOLDER ORTHONORMALIZATION OF THE SET OF VECTORS "Y".
  CALL HQR_ORTH (N, NV, Y, SIZE(Y,1), R, SIZE(R,1), &
                 RANK, PERMU, EPS_HQR)
  !====================
  ! THE MAGNITUDES OF PIVOTS IN H-QR DECOMPOSITION ARE STORED TO "R_DIAG".
  DO  L = 1, RANK
    R_DIAG(L) = ABS(R(L,L))
  ENDDO
  R_DIAG(RANK+1:NV) = 0.0_RK
  !====================
  ! DETERMINE THE EFFECTIVE RANK BY THE RELATIVE THRESHOLD VALUE "REL_THRES".
  RANK_Y = 0
  IF (R_DIAG(1) > 0.0_RK) THEN
    DO  L = 1, RANK
      IF (R_DIAG(L) <= R_DIAG(1) * REL_THRES) EXIT
      RANK_Y = RANK_Y + 1
    ENDDO
  ENDIF
  !====================
  ! SHOW THE CALCULATED PIVOT MAGNITUDES AND THE EFFECTIVE RANK.
  IF (SHOW_HQR_PIVOT) THEN
    WRITE(      *, *) 'HQR PIVOT MAGNITUDES:'
    WRITE(FILE_66, *) 'HQR PIVOT MAGNITUDES:'
    DO  L = 1, RANK
      WRITE(*,       100) L, R_DIAG(L)
      WRITE(FILE_66, 100) L, R_DIAG(L)
      100 FORMAT(3X, 'L=', I4, ':|R(L,L)|=', ES9.1)
    ENDDO
    WRITE(      *, 200) RANK_Y
    WRITE(FILE_66, 200) RANK_Y
    200 FORMAT('RANK_Y=', I4)
  ENDIF
  !====================
  DEALLOCATE (R, PERMU)
END SUBROUTINE ORTH_HQR
