!******************************************************************************
!
! $Id: orth_misc_mp.F90,v 1.2 2010/07/10 07:24:36 murakami Exp $
!
!******************************************************************************
!==============================================================================
!
!  CALCULATE THE SELF INNER PRODUCT OF VECTORS: "S := Y^{T} * Y".
!  BY THE SYMMETRY, THE MATRIX "S" IS CALCULATED ONLY ABOVE THE DIAGONAL.
!  (NOTE: THIS CODE CAN BE REPLACED BY A LEVEL-3 BLAS ROUTINE.)
!
!==============================================================================
SUBROUTINE MAT_SELF_DOT_PRODUCT (N, NV, Y, LDY, S, LDS)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN):: N         ! SIZE OF EACH VECTOR.
INTEGER, INTENT(IN):: NV        ! NUMBER OF VECTORS.
REAL(RK),INTENT(IN):: Y(LDY,NV) ! SET OF "NV" VECTORS.
INTEGER, INTENT(IN):: LDY       ! LEADING DIMENSION OF "Y".
REAL(RK),INTENT(OUT):: S(LDS,NV) ! INNER PRODUCT OF "Y" (ONLY ABOVE DIAGONALS).
INTEGER, INTENT(IN):: LDS        ! LEADING DIMENSION OF "S".
!------------------------------------------------------------------------------
INTEGER,PARAMETER:: ISTEP=128 ! THIS BLOCK SIZE IS SYSTEM DEPENDENT.
REAL(RK),ALLOCATABLE:: SPART(:,:)
INTEGER IFLAG ! STATUS FLAG FOR ALLOCATION.
INTEGER IS, IE
INTEGER I, J
!----------------------------------------------------------
#if 0
!$OMP  PARALLEL WORKSHARE
   S(:NV,:NV) = MATMUL(TRANSPOSE(Y(:N,:NV)), Y(:N,:NV))
!$OMP  END PARALLEL WORKSHARE
#else
!$OMP  PARALLEL PRIVATE(SPART,IFLAG,IE)
   ALLOCATE(SPART(NV,NV), STAT=IFLAG)
     IF(IFLAG/=0) STOP 'MAT_SELF_DOT_PRODUCT: FAILED TO ALLOCATE SPART.'
!$OMP DO
   DO  J = 1, NV
     S(:J,J) = 0.0_RK
   ENDDO
!$OMP END DO
   DO  J = 1, NV
     SPART(:J,J) = 0.0_RK
   ENDDO
!$OMP  DO
   DO  IS = 1, N, ISTEP
     IE = MIN(IS+ISTEP-1,N)
     DO  J = 1, NV
     DO  I = 1, J
       SPART(I,J) = SPART(I,J) + DOT_PRODUCT(Y(IS:IE,I), Y(IS:IE,J))
     ENDDO
     ENDDO
   ENDDO
!$OMP  END DO NOWAIT
!$OMP  CRITICAL
   DO  J = 1, NV
     S(:J,J) = S(:J,J) + SPART(:J,J)
   ENDDO
!$OMP  END CRITICAL
   DEALLOCATE (SPART)
!$OMP  END PARALLEL
#endif
END SUBROUTINE MAT_SELF_DOT_PRODUCT
!==============================================================================
!
!  TRANSFORM THE SET OF VECTORS "Y" BY A MATRIX "U" FROM RIGHT AS "Y := Y * U".
!  (NOTE: THIS ROUTINE CAN BE REPLACE BY A LEVEL-3 BLAS ROUTINE.)
!
!==============================================================================
SUBROUTINE RIGHT_TRFM (N, NV, Y, LDY, U, LDU)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
INTEGER, INTENT(IN)   :: N         ! SIZE OF VECTORS.
INTEGER, INTENT(IN)   :: NV        ! NUMBER OF VECTORS.
REAL(RK),INTENT(INOUT):: Y(LDY,NV) ! SET OF VECTORS "Y(N,NV)".
INTEGER, INTENT(IN)   :: LDY       ! LEADING DIMENSION OF "Y".
REAL(RK),INTENT(IN)   :: U(LDU,NV) ! MATRIX "U(NV,NV)".
INTEGER, INTENT(IN)   :: LDU       ! LEADING DIMENSION OF "U".
!------------------------------------------------------------------------------
INTEGER,PARAMETER:: ISTEP=128 ! THIS BLOCK SIZE IS SYSTEM DEPENDENT.
INTEGER IS, IE
!----------------------------------------------------------
#if 0
!$OMP  PARALLEL WORKSHARE
   Y(:N,:NV) = MATMUL(Y(:N,:NV), U(:NV,:NV))
!$OMP  END PARALLEL WORKSHARE
#else
!$OMP  PARALLEL DO PRIVATE(IE)
   DO  IS = 1, N, ISTEP
     IE = MIN(IS+ISTEP-1,N)
     Y(IS:IE,:NV) = MATMUL(Y(IS:IE,:NV), U(:NV,:NV))
   ENDDO
!$OMP  END PARALLEL DO
#endif
END SUBROUTINE RIGHT_TRFM
!==============================================================================
!
!  MAKE THE SMALL MATRIX "S" WHICH IS A PROJECTION OF THE SYMMETRIC MATRIX 
!  "MAT" ONTO "Y" THE SET OF BASIS OF THE SUBSPACE AS "S := Y^{T} * MAT * Y".
!  NOTE: BY THE SYMMETRY, THE MATRIX "S" IS CALCULATED ONLY ABOVE THE DIAGONALS.
!  THIS CODE IS SLOW, SINCE "Y" IS SCANNED "NV/2" TIMES IN AVERAGE.
! (THE USE OF LEVEL-3 BLAS WOULD BE FASTER.)
!
!  NOTE: THIS CODE IS NOT USED.
!==============================================================================
SUBROUTINE MAKE_PROJECTED_MAT_0 (BAND_MAT, N, NLD, NB, NV, Y, NS, S)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_MAT(NB,N) ! STORE OF BANDED SYMMETRIC MATRIX "MAT".
INTEGER, INTENT(IN):: N         ! SIZE OF MATRIX "MAT".
INTEGER, INTENT(IN):: NLD       ! LOWER BANDWIDTH OF "MAT".
INTEGER, INTENT(IN):: NB        ! LEADING DIMENSION OF "BAND_MAT".
INTEGER, INTENT(IN):: NV        ! NUMBER OF BASIS VECTORS IN "Y".
REAL(RK),INTENT(IN):: Y(N,NV)   ! SET OF BASIS VECTORS "Y" FOR PROJECTION.
INTEGER, INTENT(IN):: NS        ! LEADING DIMENSION OF "S".
REAL(RK),INTENT(OUT):: S(NS,NV) ! PROJECTED MATRIX "S" (ONLY ABOVE DIAGONALS).
!------------------------------------------------------------------------------
REAL(RK),ALLOCATABLE:: W(:)
INTEGER IFLAG
INTEGER I, J
!----------------------------------------------------------
  IF (NB < NLD) STOP 'SUBR. MAKE_PROJECTED_MAT_0: ERROR. NB < NLD.'
  IF (NS < NV)  STOP 'SUBR. MAKE_PROJECTED_MAT_0: ERROR. NS < NV.'
  ALLOCATE (W(N), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'SUBR. MAKE_PROJECTED_MAT_0: FAILED TO ALLOCATE W.'
  !====================
  DO  J = 1, NV
    ! COMPUTE "W := MAT * Y_J".
    CALL BAND_MULT (BAND_MAT, N, NLD, SIZE(BAND_MAT,1), Y(1,J), W)
    DO  I = 1, J
      S(I,J) = DOT_PRODUCT(Y(:N,I), W)
    ENDDO
  ENDDO
  !====================
  DEALLOCATE (W)
END SUBROUTINE MAKE_PROJECTED_MAT_0
!==============================================================================
!
!  MAKE THE SMALL MATRIX "S" WHICH IS A PROJECTION OF THE SYMMETRIC MATRIX 
!  "MAT" ONTO "Y" THE SET OF BASIS OF THE SUBSPACE AS "S := Y^{T} * MAT * Y".
!  NOTE: BY THE SYMMETRY, THE MATRIX "S" IS CALCULATED ONLY ABOVE THE DIAGONALS.
!  (THE USE OF LEVEL-3 BLAS WOULD BE FASTER.)
!
!==============================================================================
SUBROUTINE MAKE_PROJECTED_MAT (BAND_MAT, N, NLD, NB, NV, Y, NS, S)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_MAT(NB,N) ! STORE OF BANDED SYMMETRIC MATRIX "MAT".
INTEGER, INTENT(IN):: N         ! SIZE OF MATRIX "MAT".
INTEGER, INTENT(IN):: NLD       ! LOWER BANDWIDTH OF "MAT".
INTEGER, INTENT(IN):: NB        ! LEADING DIMENSION OF "BAND_MAT".
INTEGER, INTENT(IN):: NV        ! NUMBER OF BASIS VECTORS IN "Y".
REAL(RK),INTENT(IN):: Y(N,NV)   ! SET OF BASIS VECTORS "Y" FOR PROJECTION.
INTEGER, INTENT(IN):: NS        ! LEADING DIMENSION OF "S".
REAL(RK),INTENT(OUT):: S(NS,NV) ! PROJECTED MATRIX "S" (ONLY ABOVE DIAGONALS).
!------------------------------------------------------------------------------
INTEGER,PARAMETER:: ISTEP=128 ! THIS BLOCK SIZE IS SYSTEM DEPENDENT.
REAL(RK),ALLOCATABLE:: W(:,:)
REAL(RK),ALLOCATABLE:: SPART(:,:)
INTEGER IFLAG ! STATUS FLAG FOR ALLOCATION.
INTEGER I, J, IS, IE
!----------------------------------------------------------
   IF (NB < NLD) STOP 'SUBR. MAKE_PROJECTED_MAT: ERROR. NB < NLD.'
   IF (NS < NV)  STOP 'SUBR. MAKE_PROJECTED_MAT: ERROR. NS < NV.'
   ALLOCATE (W(N,NV), STAT=IFLAG)
     IF (IFLAG/=0) STOP 'SUBR. MAKE_PROJECTED_MAT: FAILED TO ALLOCATE W.'
   !====================
   ! COMPUTE "W := MAT * Y".
   CALL BAND_MULT_NV (BAND_MAT, N, NLD, SIZE(BAND_MAT,1), NV, Y, W)
   !====================
   ! CALCULATION OF THE INNER-PRODUCT "S := Y^{T} * W".
   ! "S" IS CALCULATED ONLY ABOVE THE DIAGONALS, BY ITS SYMMETRY.
   ! (THE USE OF LEVEL-3 BLAS WOULD BE FASTER.)
!$OMP  PARALLEL PRIVATE(SPART,IFLAG,IE)
   ALLOCATE (SPART(NV,NV), STAT=IFLAG)
     IF (IFLAG/=0) STOP 'MAKE_PROJECTED_MAT: FAILED TO ALLOCATE SPART.'
!$OMP  DO
   DO  J = 1, NV
     S(:J,J) = 0.0_RK
   ENDDO
!$OMP  END DO
   DO  J = 1, NV
     SPART(:J,J) = 0.0_RK
   ENDDO
!$OMP  DO
   DO  IS = 1, N, ISTEP
     IE = MIN(IS+ISTEP-1,N)
     DO  J = 1, NV
     DO  I = 1, J
       SPART(I,J) = SPART(I,J) + DOT_PRODUCT(Y(IS:IE,I), W(IS:IE,J))
     ENDDO
     ENDDO
   ENDDO
!$OMP  END DO NOWAIT
!$OMP  CRITICAL
   DO  J = 1, NV
     S(:J,J) = S(:J,J) + SPART(:J,J)
   ENDDO
!$OMP  END CRITICAL
   DEALLOCATE (SPART)
!$OMP  END PARALLEL
   DEALLOCATE (W)
END SUBROUTINE MAKE_PROJECTED_MAT
