!******************************************************************************
!
! $Id: ritz.F90,v 1.2 2010/07/10 07:24:36 murakami Exp $
!
!******************************************************************************
!==============================================================================
!
!  APPLY THE SIMPLE RAYLEIGH-RITZ METHOD.
!  FOR THE CASE OF THE GENERALIZED EIGENPROBLEM, VECTORS IN "Y" ARE ALREADY
!  "B"-ORTHONORMALIZED IS ASSUMED.
!
!==============================================================================
SUBROUTINE RITZ (BAND_A, N, NLD, NA, NV, Y, Q, LAM0, LAM1, &
                 MAGIN, MAGOUT)
IMPLICIT NONE
INTEGER,PARAMETER:: RK = MY_RKIND
REAL(RK),INTENT(IN):: BAND_A(NA,N) ! BANDED SYMMETRIC MATRIX "A".
INTEGER, INTENT(IN):: N            ! SIZE OF MATRICES AND VECTORS.
INTEGER, INTENT(IN):: NLD          ! LOWER BANDWIDTH OF MATRIX "A".
INTEGER, INTENT(IN):: NA           ! LEADING DIMENSION OF "BAND_A".
INTEGER, INTENT(IN):: NV         ! NUMBER OF VECTORS.
REAL(RK),INTENT(INOUT):: Y(N,NV) ! BASIS VECTORS OR APPROXIMATED EIGENVECTORS.
REAL(RK),INTENT(OUT)  :: Q(NV)   ! APPROXIMATED EIGENVALUES.
REAL(RK),INTENT(IN):: LAM0, LAM1 ! BOTH ENDS OF THE INTERVAL OF EIGENVALUE.
!---
REAL(RK),INTENT(IN) :: MAGIN(NV)  ! INPUT ROUND-OFF MAGNIFICATION RATES.
REAL(RK),INTENT(OUT):: MAGOUT(NV) ! OUTPUT ROUND-OFF MAGNIFICATION RATES.
!-----------------------------------------------------------------------
INTEGER,PARAMETER:: MX_LP=2 ! TWICE FOR ITERATION SEEMS ENOUGH.
INTEGER,PARAMETER:: FILE_66=66
REAL(RK),ALLOCATABLE:: S(:,:), U(:,:), W(:,:), MAG(:)
INTEGER IFLAG
INTEGER LP, L, NROT
LOGICAL PASS
!-----------------------------------------------------------------------
  ALLOCATE (S(NV,NV), U(NV,NV), W(NV,NV), MAG(NV), STAT=IFLAG)
    IF (IFLAG/=0) STOP 'RITZ: FAILED TO ALLOCATE S, U, W, MAG.'
  !--------
  MAG(:NV) = MAGIN(:NV)
  LOOP: DO  LP = 1, MX_LP
    !====================
    ! CALCULATE THE PROJECTED SMALL MATRIX: "S := Y^{T} * A * Y".
    CALL MAKE_PROJECTED_MAT (BAND_A, N, NLD, SIZE(BAND_A,1), &
                         NV, Y, SIZE(S,1), S)
    !====================
    ! EIGENVALUE DECOMPOSITION OF "S" TO MAKE APPROXIMATED EIGENVALUES
    ! AND THE ORTHOGONAL ROTATION MATRIX "U".
    CALL RSYM_JACOBI (NV, S, SIZE(S,1), Q, U, SIZE(U,1), NROT)
#if 1
    PRINT 100, 'RITZ LP=', LP, ', RSYM_JACOBI NROT=', NROT
    100 FORMAT (A, I1, A, I10)
#endif
    ! EIGENPAIRS OF "S" ARE SORTED BY THEIR EIGENVALUES IN ASCENDING ORDER.
    CALL EIG_SORT (NV, Q, U, SIZE(U,1), W, SIZE(W,1), 'ASCENDING')
    !====================
    ! THE SET OF BASIS VECTORS ARE ROTATED BY THE ORTHOGONAL MATRIX "U".
#if 0
    Y(:N,:NV) = MATMUL(Y(:N,:NV), U(:NV,:NV))
#else
    CALL RIGHT_TRFM (N, NV, Y, SIZE(Y,1), U, SIZE(U,1))
#endif
    !====================
    ! THE MAGNIFICATION RATES OF ROUND-OFF ERRORS ARE UPDATED.
    DO  L = 1, NV
      MAGOUT(L) = DOT_PRODUCT(MAG(:NV), ABS(U(:NV,L)))
    ENDDO
    IF (LP /= MX_LP) MAG(:NV) = MAGOUT(:NV)
    !====================
  ENDDO LOOP
  DEALLOCATE (S, U, W)
  !
#define DEBUG
#ifdef DEBUG
  !====================
  ! SHOW THE APPROXIMATED EIGENVALUES, WHETHER THEY ARE IN THE INTERVAL,
  ! THE ROUND-OFF ERROR MAGNIFICATION RATE, AND THE COUNT OF THE
  ! APPROXIMATED EIGENVALUES IN THE INTERVAL.
  WRITE(      *, *) 'Q='
  WRITE(FILE_66, *) 'Q='
  DO  L = 1, NV
    PASS = (LAM0 <= Q(L)) .AND. (Q(L) <= LAM1)
    WRITE (      *,200) L, Q(L), PASS, MAGOUT(L)
    WRITE (FILE_66,200) L, Q(L), PASS, MAGOUT(L)
    200 FORMAT('Q(',I5,')=',ES24.16,':',L,', ERROR_MAG_RATE=', ES8.1)
  ENDDO
  WRITE(      *, *)'## ', COUNT(LAM0<=Q(:).AND.Q(:)<=LAM1),&
                   'SOLUTIONS FOUND.'
  WRITE(FILE_66, *)'## ', COUNT(LAM0<=Q(:).AND.Q(:)<=LAM1),&
                   'SOLUTIONS FOUND.'
  !====================
#endif
END SUBROUTINE RITZ
